from typing import List, Optional, Dict, Iterable
import logging

from tamr_unify_client import Client
from tamr_unify_client.dataset.resource import Dataset
from tamr_toolbox.models import attribute_type

from tamr_toolbox.models.data_type import JsonDict

LOGGER = logging.getLogger(__name__)


def exists(*, client: Client, dataset_name: str) -> bool:
    """
    Check if a dataset exists in a Tamr instance

    Args:
        client: Tamr python client object for the target instance
        dataset_name: The dataset name

    Return:
        True or False for if the dataset exists in target instance
    """

    try:
        client.datasets.by_name(dataset_name)
    except KeyError:
        return False

    return True


def create(
    *,
    client: Client,
    dataset_name: str,
    dataset: Optional[Dataset] = None,
    primary_keys: Optional[List[str]] = None,
    attributes: Optional[Iterable[str]] = None,
    attribute_types: Optional[Dict[str, attribute_type.AttributeType]] = None,
    attribute_descriptions: Optional[Dict[str, str]] = None,
    description: Optional[str] = None,
    external_id: Optional[str] = None,
    tags: Optional[List[str]] = None,
) -> Dataset:
    """Flexibly create a source dataset in Tamr

    A template dataset object can be passed in to create a duplicate dataset with a new name. If
    the template dataset is not provided, the primary_keys must be defined for the dataset to be
    created. Additional attributes can be added in the attributes argument. The default attribute
    type will be ARRAY STRING. Non-default attribute types can be specified in the attribute_types
    dictionary. Any attribute descriptions can be specified in the attribute_descriptions
    dictionary.

    Args:
        client: TUC client object
        dataset_name: name for the new dataset being created
        dataset: optional dataset TUC object to use as a template for the new dataset
        primary_keys: one or more attributes for primary key(s) of the new dataset
        attributes: a list of attribute names to create in the new dataset
        attribute_types: dictionary for non-default types, attribute name is the key and
            AttributeType is the value
        attribute_descriptions: dictionary for attribute descriptions, attribute name is the key
            and the attribute description is the value
        description: description of the new dataset
        external_id: external_id for dataset, if None Tamr will create one for you
        tags: the list of tags for the new dataset

    Returns:
        Dataset created in Tamr

    Raises:
        requests.HTTPError: If any HTTP error is encountered
        ValueError: If both dataset and primary_keys are not defined
        ValueError: If the dataset already exists
        TypeError: If the attributes argument is not an Iterable

    Example:
        >>> import tamr_toolbox as tbox
        >>> tamr_client = tbox.utils.client.create(**instance_connection_info)
        >>> tbox.dataset.manage.create(
        >>>     client=tamr_client,
        >>>     dataset_name="my_new_dataset",
        >>>     primary_keys=["unique_id"],
        >>>     attributes=["name","address"],
        >>>     description="My new dataset",
        >>> )
    """

    if not dataset and not primary_keys:
        raise ValueError(f"dataset or primary_keys must be defined")

    # Get dataset information
    if dataset:
        # Get attributes from dataset object
        attribute_types, attribute_descriptions = {}, {}
        for attr in dataset.attributes.stream():
            attribute_types[attr.name] = attribute_type.from_json(attr.type.spec().to_dict())
            attribute_descriptions[attr.name] = attr.description
        attributes = attribute_types.keys()
        # Get dataset spec information
        description = dataset.description
        tags = dataset.tags
        primary_keys = dataset.key_attribute_names

    # Check input type is correct
    if attributes and not isinstance(attributes, Iterable):
        raise TypeError("attributes arg must be an Iterable")

    if not exists(client=client, dataset_name=dataset_name):
        creation_spec = {
            "name": dataset_name,
            "description": description,
            "keyAttributeNames": primary_keys,
            "externalId": external_id,
            "tags": tags,
        }
        client.datasets.create(creation_spec)
        LOGGER.info(f"A dataset with name {dataset_name} has been created")
    else:
        raise ValueError(f"A dataset with name '{dataset_name}' already exists")

    # Get new dataset
    target_dataset = client.datasets.by_name(dataset_name)

    # Update attributes in dataset
    if attributes:
        filtered_attributes = [attr for attr in attributes if attr not in primary_keys]
        create_attributes(
            dataset=target_dataset,
            attributes=filtered_attributes,
            attribute_types=attribute_types,
            attribute_descriptions=attribute_descriptions,
        )

    return target_dataset


def update(
    dataset: Dataset,
    *,
    attributes: Optional[Iterable[str]] = None,
    attribute_types: Optional[Dict[str, attribute_type.AttributeType]] = None,
    attribute_descriptions: Optional[Dict[str, str]] = None,
    description: Optional[str] = None,
    tags: Optional[List[str]] = None,
    override_existing_types: bool = False,
) -> Dataset:
    """Flexibly update a source dataset in Tamr

    All the attributes that should exist in the dataset must be defined in the attributes argument.
    This function will add/remove attributes in the dataset until the dataset attributes matches
    the set of attributes passed in as an argument. The default attribute type will be ARRAY
    STRING . To set non-default attribute types, they must be defined in the attribute_types
    dictionary. Any attribute descriptions can be specified in the attribute_descriptions
    dictionary. By default, the existing attribute types will not change unless
    override_existing_types is set to True. When False, the attribute type updates will only be
    logged.

    Args:
        dataset: An existing TUC dataset
        attributes: Complete list of attribute names that should exist in the updated dataset
        attribute_types: dictionary for non-default types, attribute name is the key and
            AttributeType is the value
        attribute_descriptions: dictionary for attribute descriptions, attribute name is the
            key and the attribute description is the value
        description: updated description of dataset, if None will not update the description
        tags: updated tags for the dataset, if None will not update tags
        override_existing_types: boolean flag, when true will alter existing attribute's types

    Returns:
        Updated Dataset

    Raises:
        requests.HTTPError: If any HTTP error is encountered
        ValueError: If the dataset is not a source dataset
        TypeError: If the attributes argument is not an Iterable

    Example:
        >>> import tamr_toolbox as tbox
        >>> from tbox.models import attribute_type
        >>> tamr_client = tbox.utils.client.create(**instance_connection_info)
        >>> dataset = = tamr_client.datasets.by_name("my_dataset_name")
        >>> tbox.dataset.manage.update(
        >>>     client=tamr_client,
        >>>     dataset=dataset,
        >>>     attributes=["unique_id","name","address","total_sales"],
        >>>     attribute_types={"total_sales":attribute_type.ARRAY(attribute_type.DOUBLE)},
        >>>     override_existing_types = True,
        >>> )
    """
    dataset_name = dataset.name
    if dataset.upstream_datasets():
        raise ValueError(f"{dataset_name} is not a source dataset")
    primary_keys = dataset.spec().to_dict()["keyAttributeNames"]

    # Check input type is correct
    if attributes and not isinstance(attributes, Iterable):
        raise TypeError("attributes arg must be an Iterable")

    # Update description and tags
    dataset_spec = dataset.spec()
    if description:
        dataset_spec = dataset_spec.with_description(description)
        LOGGER.info(f"Updating description for {dataset_name}")
    if tags:
        dataset_spec = dataset_spec.with_tags(tags)
        LOGGER.info(f"Updating tags for {dataset_name}")

    dataset_spec.put()

    if attributes:
        # Get current dataset attributes
        existing_attributes = [attr.name for attr in dataset.attributes]

        # Update attributes in dataset
        for attribute_name in attributes:
            if attribute_name in primary_keys:
                continue
            elif attribute_name in existing_attributes:
                # This attribute already exists, update to new type
                type_dict = {
                    attribute_name: (attribute_types or dict()).get(
                        attribute_name, attribute_type.DEFAULT
                    )
                }
                desc_dict = {
                    attribute_name: (attribute_descriptions or dict()).get(attribute_name)
                }

                edit_attributes(
                    dataset=dataset,
                    attribute_types=type_dict,
                    attribute_descriptions=desc_dict,
                    override_existing_types=override_existing_types,
                )
            else:
                # This attribute does not already exist, create
                create_attributes(
                    dataset=dataset,
                    attributes=[attribute_name],
                    attribute_types=attribute_types,
                    attribute_descriptions=attribute_descriptions,
                )

        # Remove any attributes from dataset that aren't in the new list of attributes
        for attribute_name in existing_attributes:
            if attribute_name not in attributes and attribute_name not in primary_keys:
                delete_attributes(dataset=dataset, attributes=[attribute_name])

    return dataset


def create_attributes(
    *,
    dataset: Dataset,
    attributes: Iterable[str],
    attribute_types: Optional[Dict[str, attribute_type.AttributeType]] = None,
    attribute_descriptions: Optional[Dict[str, str]] = None,
) -> Dataset:
    """Create new attributes in a dataset

    The default attribute type will be ARRAY STRING. To set non-default attribute types, they must
    be defined in the attribute_types dictionary. Any attribute descriptions can be specified in
    the attribute_descriptions dictionary.

    Args:
        dataset: An existing TUC dataset
        attributes: list of attribute names to be added to dataset
        attribute_types: dictionary for non-default types, attribute name is the key and
            AttributeType is the value
        attribute_descriptions: dictionary for attribute descriptions, attribute name is the
            key and the attribute description is the value

    Returns:
        Updated Dataset

    Raises:
        requests.HTTPError: If any HTTP error is encountered
        TypeError: If the attributes argument is not an Iterable
        ValueError: If the dataset is a unified dataset
        ValueError: If an attribute passed in already exists in the dataset

    """
    dataset_name = dataset.name
    if dataset.upstream_datasets():
        raise ValueError(f"{dataset_name} is not a source dataset")

    # Check input type is correct
    if not isinstance(attributes, Iterable):
        raise TypeError("attributes arg must be an Iterable")

    # Get current dataset attributes
    existing_attributes = [attr.name for attr in dataset.attributes]

    # Check that none of the new attribute names already exist
    for attribute_name in attributes:
        if attribute_name in existing_attributes:
            # This attribute already exists
            raise ValueError(
                f"An attribute with name '{attribute_name}' already exists in {dataset_name}"
            )

    # Add attributes to dataset
    for attribute_name in attributes:
        attr_spec_dict = _make_spec_dict(
            attribute_name=attribute_name,
            attribute_types=attribute_types,
            attribute_descriptions=attribute_descriptions,
        )
        dataset.attributes.create(attr_spec_dict)
        LOGGER.info(f"Created attribute '{attribute_name}' in {dataset_name}")

    return dataset


def edit_attributes(
    *,
    dataset: Dataset,
    attribute_types: Optional[Dict[str, attribute_type.AttributeType]] = None,
    attribute_descriptions: Optional[Dict[str, str]] = None,
    override_existing_types: bool = True,
) -> Dataset:
    """Edit existing attributes in a dataset

    The attribute type and/or descriptions can be updated to new values. Attributes that will be
    updated must be in either the attribute_types or attribute_descriptions dictionaries or
    both. The default attribute type will be ARRAY STRING. To set non-default attribute types, they
    must be defined in the attribute_types dictionary. Any attribute descriptions can be specified
    in the attribute_descriptions dictionary. If only the attribute_descriptions dictionary is
    defined, the attribute type will not be updated.

    Args:
        dataset: An existing TUC dataset
        attribute_types: dictionary for non-default types, attribute name is the key and
            AttributeType is the value
        attribute_descriptions: dictionary for attribute descriptions, attribute name is the
            key and the attribute description is the value
        override_existing_types: bool flag, when true will alter existing attributes

    Returns:
        Updated Dataset

    Raises:
        requests.HTTPError: If any HTTP error is encountered
        ValueError: If the dataset is not a source dataset
        ValueError: If a passed attribute does not exist in the dataset
        ValueError: If a passed attribute is a primary key and can't be removed
        ValueError: If there are no updates to attributes in attribute_types or
            attribute_descriptions arguments
    """
    dataset_name = dataset.name
    if dataset.upstream_datasets():
        raise ValueError(f"{dataset_name} is not a source dataset")

    # Check description or type changes are passed in
    if attribute_types is None and attribute_descriptions is None:
        raise ValueError(
            """Updates to attributes must be passed in via attribute_types
            or attribute_descriptions arguments"""
        )

    # Get list of attributes that need updating from attribute_types and
    # attribute_descriptions dictionaries
    attributes = {attr for attr in attribute_types or list()} | {
        attr for attr in attribute_descriptions or list()
    }

    # Get current dataset attributes
    target_attribute_dict = {attr.name: attr for attr in dataset.attributes}
    existing_attributes = target_attribute_dict.keys()
    primary_keys = dataset.spec().to_dict()["keyAttributeNames"]

    # Check that all of the attribute names already exist in dataset
    for attribute_name in attributes:
        if attribute_name not in existing_attributes:
            # This attribute does not exist
            raise ValueError(
                f"An attribute with name '{attribute_name}' does not exist in {dataset_name}"
            )
        elif attribute_name in primary_keys:
            # Can not edit a primary key
            raise ValueError(
                f"The attribute '{attribute_name}' is a primary key and can't be updated"
            )

    # Update attributes in dataset
    for attribute_name in attributes:
        attr_spec_dict = _make_spec_dict(
            attribute_name=attribute_name,
            attribute_types=attribute_types,
            attribute_descriptions=attribute_descriptions,
        )
        existing_attribute_spec = target_attribute_dict[attribute_name].spec()
        if attribute_types is None or attribute_name not in attribute_types:
            new_type_class = attribute_type.from_json(existing_attribute_spec.to_dict()["type"])
        else:
            new_type_class = attribute_type.from_json(attr_spec_dict["type"])
        old_type_class = attribute_type.from_json(existing_attribute_spec.to_dict()["type"])

        if new_type_class == old_type_class:
            # Update description
            if (
                attribute_descriptions is not None
                and attribute_name in attribute_descriptions.keys()
            ):
                existing_attribute_spec = existing_attribute_spec.with_description(
                    attribute_descriptions[attribute_name]
                )
                existing_attribute_spec.put()
            else:
                LOGGER.info(
                    f"There are no updates to the attribute '{attribute_name}' in {dataset_name}"
                )
        elif override_existing_types:
            # Update type
            new_attr_spec = existing_attribute_spec.to_dict()
            new_attr_spec["type"] = attr_spec_dict["type"]

            # Update description
            if "description" in attr_spec_dict.keys():
                new_attr_spec["description"] = attr_spec_dict["description"]

            # Remove and add attribute with new spec
            dataset.attributes.delete_by_resource_id(
                target_attribute_dict[attribute_name].resource_id
            )
            dataset.attributes.create(new_attr_spec)
            LOGGER.info(f"Updated attribute '{attribute_name}' in {dataset_name}")
        else:
            LOGGER.info(
                f"""The attribute '{attribute_name}' in {dataset_name} curently has
                 the type '{str(old_type_class)}'. Set 'override_existing_types' to
                 True to update the type to '{str(new_type_class)}'
                """
            )

    return dataset


def delete_attributes(*, dataset: Dataset, attributes: Iterable[str] = None) -> Dataset:
    """Remove attributes from dataset by attribute name

    Args:
        dataset: An existing TUC dataset
        attributes: list of attribute names to delete from dataset

    Returns:
        Updated Dataset

    Raises:
        ValueError: If the dataset is not a source dataset
        ValueError: If a passed attribute does not exist in the dataset
        ValueError: If a passed attribute is a primary key and can't be removed
        TypeError: If the attributes argument is not an Iterable
    """
    dataset_name = dataset.name
    if dataset.upstream_datasets():
        raise ValueError(f"{dataset_name} is not a source dataset")

    # Check input type is correct
    if not isinstance(attributes, Iterable):
        raise TypeError("attributes arg must be an Iterable")

    # Get current dataset attributes
    target_attribute_dict = {attr.name: attr for attr in dataset.attributes}
    existing_attributes = target_attribute_dict.keys()
    primary_keys = dataset.spec().to_dict()["keyAttributeNames"]

    # Check all attributes exist before starting to remove any
    for attribute_name in attributes:
        if attribute_name not in existing_attributes:
            raise ValueError(f"The attribute '{attribute_name}' does not exist in {dataset_name}")
        elif attribute_name in primary_keys:
            # Can not edit a primary key
            raise ValueError(
                f"The attribute '{attribute_name}' is a primary key and can't be removed"
            )

    # Remove attributes from dataset
    for attribute_name in attributes:
        dataset.attributes.delete_by_resource_id(target_attribute_dict[attribute_name].resource_id)
        LOGGER.info(f"Deleted attribute '{attribute_name}' in {dataset_name}")

    return dataset


def update_records(
    dataset: Dataset,
    *,
    updates: Optional[list] = None,
    delete_all: bool = False,
    primary_keys: List[str],
    primary_key_name: str,
):
    """Flexibly update the records of a dataset. The user supplies a list of primary keys for a
    subset of the datasets records, along with a list of updates describing how each record should
    be altered. An update should either be the string "delete" or a dictionary in
    "attribute: value" format. In the first case, the record having the corresponding primary key
    is deleted, and in the second case, the data in the dictionary replaces the record having the
    corresponding primary key. If no such record exists, a new record is created. Alternatively,
    the user can set a flag to specify that all records indicated by the list of primary keys
    should be deleted.

    Args:
        dataset: An existing TUC dataset
        updates: List of updates to push to the dataset
        delete_all: Whether all indicated records should be deleted
        primary_keys: List of primary key values for all target records
        primary_key_name: Name of the primary key of the target dataset

    Returns:
        Updated dataset

    Raises:
        KeyError: If an indicated attribute does not exist
        TypeError: If an update in the list is not "delete" or a dict
        ValueError: If updates and primary_keys have differing lengths
    """

    # If delete_all, create an updates list of all deletes. Otherwise, ensure that the list of
    # updates and the list of primary keys have the same length.
    if delete_all:
        updates = ["delete"] * len(primary_keys)
    else:
        if updates is None or len(primary_keys) != len(updates):
            raise ValueError(
                f"Arguments updates and primary_keys must exist and have equal length"
            )

    # Populate list of primary keys for deletion and records to upsert.
    deletions = []
    records = []
    attributes = [attribute.name for attribute in dataset.attributes]
    for i in range(len(updates)):
        if updates[i] == "delete":
            deletions.append(primary_keys[i])
        else:
            if not isinstance(updates[i], dict):
                raise TypeError(f"Invalid update at index {i}")
            if primary_key_name not in updates[i]:
                updates[i][primary_key_name] = primary_keys[i]
            for k in updates[i]:
                if k not in attributes:
                    raise KeyError(f"Key {k} is not an attribute of input dataset")
            records.append(updates[i])

    # Carry out any deletions.
    if deletions:
        dataset.delete_records_by_id(deletions)

    # Carry out any upserts.
    if records:
        dataset.upsert_records(records, primary_key_name)

    return dataset


def _make_spec_dict(
    attribute_name: str,
    attribute_types: Dict[str, attribute_type.AttributeType],
    attribute_descriptions: Dict[str, str],
) -> JsonDict:
    """Create attribute spec dictionary

    The default attribute type will be ARRAY STRING. Non-default attribute types can be specified
    in the attribute_types dictionary. Any attribute descriptions can be specified in the
    attribute_descriptions dictionary.

    Args:
        attribute_name: name of the attribute
        attribute_types: dictionary for non-default types, attribute name is the key and
            AttributeType is the value
        attribute_descriptions: dictionary for attribute descriptions, attribute name is the
            key and the attribute description is the value

    Returns:
        Json Dict
    """
    if attribute_types is not None and attribute_name in attribute_types.keys():
        attr_type = attribute_types[attribute_name]
    else:
        attr_type = attribute_type.DEFAULT

    result = {"name": attribute_name, "type": attribute_type.to_json(attr_type=attr_type)}

    if attribute_descriptions is not None and attribute_name in attribute_descriptions.keys():
        result["description"] = attribute_descriptions[attribute_name]
    return result



"""Utilities shared by enrichment services."""
import json
import os
from typing import Any, List, Optional, Tuple


# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke docs` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    import pandas as pd


class SetEncoder(json.JSONEncoder):
    """A Class to transform type 'set' to type 'list' when saving objects to JSON."""

    def default(self, python_object):
        """
        Transform a set into a list if input is a set

        Args:
            python_object: the python object to be saved to a json format

        Returns:
            Default json encoder format of input object or List if input is a Set
        """
        if isinstance(python_object, set):
            return list(python_object)
        return json.JSONEncoder.default(self, python_object)


def _yield_chunk(list_to_split: List[Any], chunk_size: int) -> List[List[Any]]:
    """
    Split a list into a List of List with constant length

    Args:
        list_to_split: List to split into chunks
        chunk_size: number of items to have in each list after splitting

    Returns:
        A List of List
    """

    # For item i in a range that is a length of l,
    for i in range(0, len(list_to_split), chunk_size):
        # Create an index range for l of n items:
        yield list_to_split[i : i + chunk_size]


def create_empty_mapping(path: str) -> str:
    """
    Create an empty mapping on disk.

    Args:
        path: location where empty mapping is created

    Returns:
        A path to the new empty file
    """
    with open(path, "w") as file:
        file.write(json.dumps({}))
    return path


def dataframe_to_tuples(
    dataframe: "pd.DataFrame", columns_to_join: List[str]
) -> List[Tuple[Optional[str], ...]]:
    """Generate list of string tuples from specified dataframe columns.

    Args:
        dataframe: a dataframe
        columns_to_join: list of columns to join in the output

    Raises:
        ValueError: if any of the `columns_to_join` do not appear in the dataframe

    Returns:
        list of strig tuples generated from specified columns of the dataframe
    """
    # Check that expexted columns exist
    df_columns = dataframe.columns
    if not all([x in df_columns for x in columns_to_join]):
        raise ValueError(
            f"Not all columns {columns_to_join} exist in input dataframe columns {df_columns}"
        )

    tuples = [
        tuple(x)
        for x in dataframe[columns_to_join]
        .astype(str)
        .replace({"nan": None, "None": None})
        .to_records(index=False)
    ]

    return tuples


def join_clean_tuple(tup: Tuple[Optional[str], ...]) -> str:
    """Join tuple entries, stripping extra leading/trailing whitespace and uppercasing.

    Args:
        tup: tuple of string or None values

    Returns:
        uppercased string made from joining tuple entries.
    """
    return " ".join([x.strip().upper() for x in tup if x])



"""Tasks related to backup and restore of Tamr instances"""
import logging
import requests
from tamr_unify_client import Client
from tamr_toolbox.filesystem import bash
from tamr_toolbox.models.data_type import JsonDict
from tamr_toolbox import utils
from typing import Generator, Union, Optional, List
from datetime import datetime
import os
import json
from pathlib import Path

LOGGER = logging.getLogger(__name__)


def list_backups(client: Client) -> Generator[JsonDict, None, None]:
    """Lists all backups available to Tamr client. Will list both succeeded and failed backups.

    Args:
        client: A client object
    Returns:
        A generator of json dict objects for the backups available to client."""
    response = client.get("backups")
    for backup in response.json():
        yield backup


def get_backup_by_id(client: Client, backup_id: str) -> JsonDict:
    """Fetches the json object for a given backup ID.

    Args:
        client: A Tamr client object.
        backup_id: The relativeID corresponding to the desired backup.

    Returns:
        Json dict corresponding to the desired backup.

    Raises:
        ValueError: Raised if GET request to Tamr fails
    """
    api_string = f"backups/{backup_id}"
    response = client.get(api_string)

    if not response.ok:
        message = (
            f"Received non-200 response code '{response.status_code}' "
            f"with message '{response.json()['message']}': '{response.json()}'"
        )
        LOGGER.error(message)
        raise ValueError(message)

    return response.json()


def initiate_backup(
    client: Client,
    *,
    poll_interval_seconds: int = 30,
    polling_timeout_seconds: Optional[int] = None,
    connection_retry_timeout_seconds: int = 600,
) -> requests.Response:
    """Runs a backup of Tamr client and waits until it is finished.

    Args:
        client: A Tamr client object
        poll_interval_seconds: Amount of time in between polls of job state.
        polling_timeout_seconds: Amount of time before a timeout error is thrown.
        connection_retry_timeout_seconds: Amount of time before timeout error is thrown during
            connection retry

    Returns:
        Json dict of response from API request."""
    response = client.post("backups")
    if not response.ok:
        message = f"Received non-200 response code '{response.status_code}': {response.json()}"
        LOGGER.error(message)
        raise RuntimeError(message)

    backup_id = response.json()["relativeId"]

    op = utils.client.poll_endpoint(
        client=client,
        api_endpoint=f"backups/{backup_id}",
        poll_interval_seconds=poll_interval_seconds,
        polling_timeout_seconds=polling_timeout_seconds,
        connection_retry_timeout_seconds=connection_retry_timeout_seconds,
    )
    return op


def initiate_restore(
    client: Client,
    backup_id: str,
    *,
    polling_timeout_seconds: Optional[int] = None,
    poll_interval_seconds: int = 30,
    connection_retry_timeout_seconds: int = 600,
) -> requests.Response:
    """Restores the Tamr client to the state of the supplied backup.

    Args:
        client: A Tamr client object
        backup_id: BackupId of the desired backup.
        polling_timeout_seconds: Amount of time before a timeout error is thrown.
        poll_interval_seconds: Amount of time in between polls of job state.
        connection_retry_timeout_seconds: Amount of time before timeout error is thrown during
            connection retry

    Returns:
        Json dict of response from API request.

    Raises:
        ValueError: Raised if the target backup contains errors
        RuntimeError: Raised if the restore fails to start
    """
    backup = get_backup_by_id(client=client, backup_id=backup_id)
    backup_state = backup["state"]
    if not backup_state == "SUCCEEDED":
        value_error_message1 = (
            f"Backup file with ID {backup_id} did not succeed and has status {backup_state}"
        )
        raise ValueError(value_error_message1)

    error_message = backup["errorMessage"]
    if not error_message == "":
        value_error_message2 = (
            f"Backup file with ID {backup_id} contains non-null error message {error_message}"
        )
        LOGGER.error(value_error_message2)
        raise ValueError(value_error_message2)

    response = client.post("instance/restore", data=backup_id)
    if not response.ok:
        runtime_error_message = (
            f"Received non-200 response code '{response.status_code}' : {response.json()}"
        )
        LOGGER.error(runtime_error_message)
        raise RuntimeError(runtime_error_message)

    op = utils.client.poll_endpoint(
        client=client,
        api_endpoint="instance/restore",
        poll_interval_seconds=poll_interval_seconds,
        polling_timeout_seconds=polling_timeout_seconds,
        connection_retry_timeout_seconds=connection_retry_timeout_seconds,
    )

    return op


def validate_backup(
    directory: Union[Path, str], *, backup_datetime_format: str = "%Y-%m-%d_%H-%M-%S-%f"
) -> bool:
    """Validates that a directory is a valid backup. A valid backup has a
    manifest file, a completion file (_SUCCEEDED, _FAILED, or _CANCELED),
    the folder has a valid date format, and the date is prior to the current time

    Args:
        directory: path to backup directory
        backup_datetime_format: String datetime format in backup folder name

    Returns:
        True if directory is a valid backup, otherwise False.

    """
    if not isinstance(directory, Path):
        directory = Path(directory)
    backup_name = directory.stem

    if not os.path.exists(directory / "manifest.json") and not os.path.exists(
        directory / "manifest.yaml"
    ):
        LOGGER.warning(
            f"The following directory '{directory}' is not a valid Tamr Backup. "
            f"Neither manifest.json nor manifest.yaml exist."
        )
        return False
    if not (
        any([os.path.exists(directory / f"_{x}") for x in ["SUCCEEDED", "CANCELED", "FAILED"]])
    ):
        LOGGER.warning(
            f"The following directory '{directory}' is not a valid Tamr Backup "
            f"A SUCCEEDED, CANCELED, or FAILED completion file does not exist."
        )
        return False

    try:
        time = datetime.strptime(backup_name, backup_datetime_format)
    except ValueError:
        LOGGER.warning(
            f"The following directory '{directory}' is not a valid Tamr Backup "
            f"The directory name {backup_name} is not a valid datetime"
        )
        return False

    if datetime.now() > time > datetime(2010, 1, 1, 0, 0, 0, 0):
        return True
    else:
        LOGGER.warning(
            f"The following directory '{directory}' is not a valid "
            f"Tamr Backup the datetime represents an invalid time {time}."
        )
        return False


def delete_backups(
    *,
    backups: List[str],
    backup_directory: Union[Path, str],
    backup_datetime_format: str = "%Y-%m-%d_%H-%M-%S-%f",
) -> List[str]:
    """Deletes backup folders recursively.

    Args:
        backups: list of backups to delete
        backup_directory: Path to backup directory
        backup_datetime_format: String datetime format in backup folder name

    Returns:
        list of deleted backup names
    """
    if not isinstance(backup_directory, Path):
        backup_directory = Path(backup_directory)
    backups_to_delete = sorted(backups, reverse=True)
    LOGGER.info(f"Deleting {len(backups_to_delete)} backup(s): '{backups_to_delete}'")

    deleted_backups = []
    for backup in backups:
        backup_path = os.path.join(backup_directory, backup)
        if validate_backup(Path(backup_path), backup_datetime_format=backup_datetime_format):
            bash.remove_directories([backup_path], allow_recursive_deletes=True)
            deleted_backups.append(backup)
        else:
            LOGGER.error(f"{backup} is not a valid backup to delete.")
    return deleted_backups


def classify_backups(
    backup_directory: Union[Path, str], *, backup_datetime_format: str = "%Y-%m-%d_%H-%M-%S-%f"
) -> JsonDict:
    """Takes stock of successful and failed valid backups in the backup directory.

    Args:
        backup_directory: Path to backup directory
        backup_datetime_format: String datetime format in backup folder name

    Returns:
        JSON dict with the keys "successful" (List of successful backups) and
        "not_successful" (List of failed or cancelled backups)

    Raises:
        ValueError: if target backup file contains an error message
    """
    if not isinstance(backup_directory, Path):
        backup_directory = Path(backup_directory)
    backups = os.listdir(backup_directory)

    LOGGER.info(f"Found '{len(backups)}' entries in the directory {str(backup_directory)}")

    not_succeeded = []
    succeeded = []
    for backup in backups:
        if validate_backup(
            backup_directory / backup, backup_datetime_format=backup_datetime_format
        ):
            backup_path = backup_directory / backup

            if os.path.exists(backup_path / "_SUCCEEDED"):
                succeeded.append(backup)
                with open(backup_path / "_SUCCEEDED") as f:
                    succeeded_file = json.load(f)
                if not succeeded_file["errorMessage"] is None:
                    raise ValueError(
                        f"Successful backup file {backup_path}/_SUCCEEDED contains non-null "
                        f"error message '{succeeded_file['errorMessage']}'"
                    )
            else:
                not_succeeded.append(backup)

    return {"succeeded": succeeded, "not_succeeded": not_succeeded}


def delete_old_backups(
    backup_directory: Union[Path, str],
    *,
    num_successful_backups_to_keep: int,
    num_failed_backups_to_keep: int,
    backup_datetime_format: str = "%Y-%m-%d_%H-%M-%S-%f",
) -> Optional[List[JsonDict]]:
    """Deletes old backups. Keeps the most recent num_successful_backups_to_keep successful backups
        and the most recent num_failed_backups_to_keep failed backups

    Args:
        backup_directory: Path to backup directory
        num_successful_backups_to_keep: Number of successful backups to keep
        num_failed_backups_to_keep: Number of failed or canceled backups to keep
        backup_datetime_format: String datetime format in backup folder name

    Returns:
        A list of deleted backups. Returns None if no backups are deleted.

    Raises:
        ValueError: if the number of backups to keep is less than 0

    """
    if not isinstance(backup_directory, Path):
        backup_directory = Path(backup_directory)
    if num_successful_backups_to_keep < 0:
        value_error_message_1 = (
            f"Argument for num_successful_backups_to_keep must be greater than 0. "
            f"Found {num_successful_backups_to_keep}"
        )
        LOGGER.error(value_error_message_1)
        raise ValueError(value_error_message_1)

    if num_failed_backups_to_keep < 0:
        value_error_message_2 = (
            f"Argument for num_failed_backups_to_keep must be greater than 0. "
            f"Found '{num_failed_backups_to_keep}'"
        )
        LOGGER.error(value_error_message_2)
        raise ValueError(value_error_message_2)

    LOGGER.info(f"Fetching all backups from directory '{backup_directory}'")
    result = classify_backups(backup_directory, backup_datetime_format=backup_datetime_format)
    succeeded = result["succeeded"]
    not_succeeded = result["not_succeeded"]
    if (len(succeeded) < 1) and (len(not_succeeded) < 1):
        LOGGER.info(f"Found zero backups in directory '{backup_directory}'")
        return None
    else:
        deleted_failed_backups = []
        deleted_successful_backups = []
        if len(succeeded) <= num_successful_backups_to_keep:
            LOGGER.info(
                f"Found {len(succeeded)} successful backup. Keeping all successful backups"
            )
        else:
            LOGGER.info(
                f"Removing {len(succeeded) - num_successful_backups_to_keep} successful backups"
            )
            succeeded = sorted(succeeded, reverse=False)
            LOGGER.warning(
                f"backups {succeeded} from directory {backup_directory} "
                f"which is {type(backup_directory)}"
            )
            deleted_successful_backups = delete_backups(
                backups=succeeded[: len(succeeded) - num_successful_backups_to_keep],
                backup_directory=backup_directory,
                backup_datetime_format=backup_datetime_format,
            )
        if len(not_succeeded) <= num_failed_backups_to_keep:
            LOGGER.info(f"Found {len(not_succeeded)} failed backups. Keeping all failed backups")
        else:
            LOGGER.info(
                f"Removing {len(not_succeeded) - num_failed_backups_to_keep} unsuccessful backups"
            )
            not_succeeded = sorted(not_succeeded, reverse=False)
            deleted_failed_backups = delete_backups(
                backups=not_succeeded[: len(not_succeeded) - num_failed_backups_to_keep],
                backup_directory=backup_directory,
                backup_datetime_format=backup_datetime_format,
            )
        return deleted_failed_backups + deleted_successful_backups


def delete_old_spark_event_logs(
    tamr_home_directory: Union[Path, str], *, num_days_to_keep: int = 14
) -> List[str]:
    """
    Deletes sparkEventLogs older than the specified number of days.
    This assumes that Spark is running locally on the same VM as Tamr
    and that the logs are on the local filesystem.

    Args:
        tamr_home_directory: Path to the Tamr home directory
        num_days_to_keep: Number of days for which to keep logs

    Returns:
        A list of deleted sparkEventLogs files

    Raises:
        ValueError: if num_days_to_keep is less than 0
        FileNotFoundError: if sparkEventLogs directory doesn't exist
    """
    spark_event_log_directory = os.path.join(
        tamr_home_directory, "tamr/unify-data/job/sparkEventLogs"
    )
    if not os.path.exists(spark_event_log_directory):
        message = f"directory does not exist: {spark_event_log_directory}"
        LOGGER.error(message)
        raise FileNotFoundError(message)

    return bash.delete_old_files(spark_event_log_directory, num_days_to_keep=num_days_to_keep)



"""Tasks related to creating, updating, saving, and moving address validation data from Tamr"""
import copy
import json
import logging
import os
from dataclasses import asdict, dataclass, fields
from typing import Dict, List, Optional, Union

from requests.exceptions import HTTPError
from tamr_unify_client.dataset.collection import DatasetCollection
from tamr_unify_client.dataset.resource import Dataset
from typing_extensions import Literal

from tamr_toolbox.enrichment.enrichment_utils import create_empty_mapping

LOGGER = logging.getLogger(__name__)


Granularity = Literal[
    "GRANULARITY_UNSPECIFIED",
    "SUB_PREMISE",
    "PREMISE",
    "PREMISE_PROMXIMITY",
    "BLOCK",
    "ROUTE",
    "OTHER",
]


@dataclass
class AddressValidationMapping:
    """DataClass for address validation data.

    Args:
        input_address: input address
        validated_formatted_address: the "formattedAddress" returns by the validation API, if any
        expiration: the expiration timestamp of the data, 30 days from API call
        region_code: region code returned by the validation API
        postal_code: postal code returned by the validation API
        admin_area: administrative area returned by the validation API (state for US addresses)
        locality: locality returned by the validation API (city/town for US addresses)
        address_lines: address lines returned by the validation API (e.g. ['66 Church St'])
        usps_firstAddressLine: first address line in validated USPS format, if available
        usps_cityStateZipAddressLine: : second address line in validated USPS format, if available
        usps_city: city in validated USPS format, if available
        usps_state: state in validated USPS format, if available
        usps_zipCode: str = "
        latitude: latitude associated with validated address, if any
        longitude: longitude associated with validated address, if any
        place_id: the google placeId -- the only result field not subject to the `expiration`
        input_granularity: granularity of input given by validation API
        validation_granularity: granularity of validation given by validation API
        geocode_granularity: granularity of geocode given by validation API
        has_inferred: whether the result has inferred components
        has_unconfirmed: whether the result has unconfirmed components
        has_replaced: whether the result has replaced components
        address_complete: whether the input was complete
    """

    input_address: str
    validated_formatted_address: Optional[str]
    expiration: str  # timestamp in the format given by `str(datetime.now())`
    region_code: Optional[str]
    postal_code: Optional[str]
    admin_area: Optional[str]
    locality: Optional[str]
    address_lines: List[str]
    usps_first_address_line: Optional[str]
    usps_city_state_zip_line: Optional[str]
    usps_city: Optional[str]
    usps_state: Optional[str]
    usps_zip_code: Optional[str]
    latitude: Optional[float]
    longitude: Optional[float]
    place_id: Optional[str]
    input_granularity: Granularity
    validation_granularity: Granularity
    geocode_granularity: Granularity
    has_inferred: bool
    has_unconfirmed: bool
    has_replaced: bool
    address_complete: bool


def to_dict(
    dictionary: Dict[str, AddressValidationMapping]
) -> List[Dict[str, Union[str, List[str], float, None]]]:
    """
    Convert a toolbox address validation mapping entries to  list-of-dictionary format.

    Args:
        dictionary: a toolbox address validation mapping

    Returns:
        A list of toolbox address validation mapping entries in dictionary format
    """
    return [asdict(t) for t in dictionary.values()]


def update(
    main_dictionary: Dict[str, AddressValidationMapping],
    tmp_dictionary: Dict[str, AddressValidationMapping],
) -> None:
    """
    Update a toolbox address validation mapping with another temporary address validation mapping

    Args:
        main_dictionary: the main toolbox address validation mapping containing prior results
        tmp_dictionary: a temporary toolbox address validation mapping containing new data
    """
    for input_addr, mapping in tmp_dictionary.items():
        main_dictionary[input_addr] = copy.copy(mapping)


def from_dataset(dataset: Dataset) -> Dict[str, AddressValidationMapping]:
    """
    Stream an address validation mapping dataset from Tamr.

    Args:
        dataset: Tamr Dataset object

    Returns:
        A toolbox address validation mapping

    Raises:
        ValueError: if the provided `dataset` is not a toolbox address validation mapping dataset
        NameError: if the provided `dataset` does not contain all the attributes of a
            toolbox address validation mapping
        RuntimeError: if there is any other problem while reading the `dataset` as a
            toolbox address validation mapping
    """
    if dataset.key_attribute_names[0] != "input_address":
        error_message = "Provided Tamr Dataset is not a toolbox address validation mapping"
        LOGGER.error(error_message)
        raise ValueError(error_message)

    dictionary = {}
    for record in dataset.records():
        try:
            # Values are returned as a length-1 list of string, we change this to strings
            entry = AddressValidationMapping(
                input_address=record["input_address"],
                validated_formatted_address=record["validated_formatted_address"][0],
                expiration=record["expiration"][0],
                region_code=record["region_code"][0] if record["region_code"] else None,
                postal_code=record["postal_code"][0] if record["postal_code"] else None,
                admin_area=record["admin_area"][0] if record["admin_area"] else None,
                locality=record["locality"][0] if record["locality"] else None,
                address_lines=record["address_lines"] if record["address_lines"] else [],
                usps_first_address_line=record["usps_first_address_line"]
                if record["usps_first_address_line"]
                else None,
                usps_city_state_zip_line=record["usps_city_state_zip_line"]
                if record["usps_city_state_zip_line"]
                else None,
                usps_city=record["usps_city"] if record["usps_city"] else None,
                usps_state=record["usps_state"] if record["usps_state"] else None,
                usps_zip_code=record["usps_zip_code"] if record["usps_zip_code"] else None,
                latitude=float(record["latitude"][0]) if record["latitude"] else None,
                longitude=float(record["longitude"][0]) if record["longitude"] else None,
                place_id=record["place_id"][0] if record["place_id"] else None,
                input_granularity=record["input_granularity"][0]
                if record["input_granularity"]
                else "GRANULARITY_UNSPECIFIED",
                validation_granularity=record["validation_granularity"][0]
                if record["validation_granularity"]
                else "GRANULARITY_UNSPECIFIED",
                geocode_granularity=record["geocode_granularity"][0]
                if record["geocode_granularity"]
                else "GRANULARITY_UNSPECIFIED",
                has_inferred=record["has_inferred"][0] if record["has_inferred"] else False,
                has_unconfirmed=record["has_unconfirmed"][0]
                if record["has_unconfirmed"]
                else False,
                has_replaced=record["has_replaced"][0] if record["has_replaced"] else False,
                address_complete=record["address_complete"][0]
                if record["address_complete"]
                else False,
            )

        except KeyError as exp:
            error_message = (
                f"Supplied Tamr dataset is not in toolbox address validation mapping format: {exp}"
            )
            LOGGER.error(error_message)
            raise NameError(error_message) from exp
        except Exception as exp:
            error_message = f"Error while reading Tamr dataset address validation mapping: {exp}"
            LOGGER.error(error_message)
            raise RuntimeError(error_message) from exp

        dictionary.update({entry.input_address: entry})
    return dictionary


def to_dataset(
    addr_mapping: Dict[str, AddressValidationMapping],
    *,
    dataset: Optional[Dataset] = None,
    datasets_collection: Optional[DatasetCollection] = None,
    create_dataset: bool = False,
    dataset_name: str = "address_validation_mapping",
) -> str:
    """Ingest a toolbox address validation mapping in Tamr, creating the source dataset if needed.

    Args:
        addr_mapping: a toolbox address validation mapping
        dataset: a Tamr client dataset
        datasets_collection: a Tamr client datasets collection
        create_dataset: flag to create or upsert to an existing address validation mapping
            source dataset
        dataset_name: name to use if creating new dataset

    Returns:
        The name of the created or updated Tamr Dataset

    Raises:
        ValueError: if `create_dataset` is False and `dataset` is not provided or is not a
            toolbox address validation mapping dataset.
            If `create_dataset` is True but `datasets_collection` or `target_language` or
            `source_language` is missing or the Tamr dataset already exists
        RuntimeError: if there is an error during the creation of the Tamr dataset attributes
    """
    if create_dataset is False:
        if dataset is None:
            error_message = (
                "Tamr Client Dataset missing from inputs. Please provide a Tamr "
                "Client Dataset if updating an existing address validation dataset"
            )
            LOGGER.error(error_message)
            raise ValueError(error_message)

        if dataset.key_attribute_names[0] != "input_address":
            error_message = "Provided Tamr Dataset is not a toolbox address validation mapping"
            LOGGER.error(error_message)
            raise ValueError(error_message)

    else:
        if not datasets_collection:
            error_message = (
                "Tamr Datasets Collection must be specified to create address validation dataset."
            )
            LOGGER.error(error_message)
            raise ValueError(error_message)

        if dataset_name in [d.name for d in datasets_collection]:
            error_message = (
                f"Tamr Dataset {dataset_name} already exists on Tamr, you cannot duplicate it."
            )
            LOGGER.error(error_message)
            raise ValueError(error_message)

        LOGGER.info("Creating toolbox address validation dataset %s on Tamr", dataset_name)
        creation_spec = {"name": dataset_name, "keyAttributeNames": ["input_address"]}
        dataset = datasets_collection.create(creation_spec)

        attributes = dataset.attributes
        for attribute in [att.name for att in fields(AddressValidationMapping)]:
            if attribute == "input_address":
                continue

            attr_spec = {
                "name": attribute,
                "type": {"baseType": "ARRAY", "innerType": {"baseType": "STRING"}},
            }
            try:
                attributes.create(attr_spec)
            except HTTPError as exp:
                error_message = (
                    f"Error while creating attribute {attribute} for dataset {dataset_name}: {exp}"
                )
                LOGGER.error(error_message)
                raise RuntimeError(error_message) from exp

    LOGGER.info("Ingesting toolbox address validation mapping to Tamr")
    dataset.upsert_records(records=to_dict(addr_mapping), primary_key_name="input_address")
    return dataset.name


def to_json(dictionary: Dict[str, AddressValidationMapping]) -> List[str]:
    """
    Convert a toolbox address validation mapping entries to a json format where set object are
    converted to list

    Args:
        dictionary: a toolbox address validation mapping

    Returns:
        A list of toolbox address validation mapping entries in json format
    """
    return [json.dumps(asdict(t)) for t in dictionary.values()]


def save(
    addr_mapping: Dict[str, AddressValidationMapping],
    addr_folder: str,
    filename: str = "address_validation_mapping.json",
) -> None:
    """
    Save a toolbox address validation mapping to disk

    Args:
        addr_mapping: dictionary object to be saved to disk
        addr_folder: base directory where mapping is saved
        filename: filename to use to save
    """
    addr_filepath = os.path.join(addr_folder, filename)

    if len(addr_mapping) > 0:
        LOGGER.debug("Writing address mapping to file")
        with open(addr_filepath, "w") as f:
            f.write("\n".join(to_json(addr_mapping)))


def load(
    addr_folder: str, filename: str = "address_validation_mapping.json"
) -> Dict[str, AddressValidationMapping]:
    """
    Load a toolbox address validation mapping from disk to memory

    Args:
        addr_folder: base directory where mapping is saved
        filename: filename where mapping is saved

    Returns:
        A toolbox address validation mapping

    Raises:
        RuntimeError: if the file was found on disk but is not of a valid toolbox address
          validation mapping type
    """
    filepath = os.path.join(addr_folder, filename)

    if not os.path.exists(filepath):
        LOGGER.info("Dictionary %s does not exist, creating an empty one.", filepath)
        filepath = create_empty_mapping(path=filepath)
        return {}

    with open(filepath, "r") as f:
        mapping_lst = [json.loads(line) for line in f.readlines()]
        try:
            # Tranform the loaded dictionaries into a AddressValidationMapping
            mapping_lst = [AddressValidationMapping(**t) for t in mapping_lst if t]
            # Make the standardized phrase the main key of the address validation mapping
            mapping_dict = {t.input_address: t for t in mapping_lst}
        except Exception as excp:
            error_message = (
                f"Could not read address validation mapping at {filepath}. "
                f"Check that the dictionary is of the correct type. Error: {excp}"
            )
            LOGGER.error(error_message)
            raise RuntimeError(error_message) from excp

    return mapping_dict



"""Tasks related to moving data in or out of Tamr using delimited files"""
from typing import Optional, List, Union, Dict
from functools import partial
from pathlib import Path

import csv
import logging
import os
import requests

from tamr_unify_client.dataset.resource import Dataset
from tamr_unify_client.project.resource import Project

from tamr_toolbox.utils import operation
from tamr_toolbox.data_io import common

LOGGER = logging.getLogger(__name__)


def from_dataset(
    dataset: Dataset,
    export_file_path: Union[Path, str],
    *,
    csv_delimiter: str = ",",
    columns: Optional[List[str]] = None,
    column_name_dict: Optional[Dict[str, str]] = None,
    flatten_delimiter: str = "|",
    quote_character: str = '"',
    quoting: int = csv.QUOTE_MINIMAL,
    na_value: str = "NaN",
    nrows: Optional[int] = None,
    allow_dataset_refresh: bool = False,
    buffer_size: int = 10000,
    overwrite: bool = False,
    encoding: str = "utf-8",
) -> int:
    """
    Export a Tamr Dataset to a csv file. Records are streamed to disk and written according to a
    given buffer size. As a result this is more memory efficient than first reading to a
    pandas.DataFrame and writing to CSV.

    Args:
        dataset: Tamr Dataset object
        export_file_path: Path to the csv file where the dataset will be saved
        csv_delimiter: Delimiter of the csv file
        columns: Optional, Ordered list of columns to write. If None, write all columns in
            arbitrary order.
        column_name_dict: Optional, Dictionary in the format {<Tamr dataset column name> : <new csv
            column name>}, used to rename some or all columns in the output file.
        flatten_delimiter: Flatten list types to strings by concatenating with this delimiter
        quote_character: Character used to escape value for csv delimiter when it appears in the
            value.
        quoting: The escape strategy to use according to the Python csv writer.
            See https://docs.python.org/2/library/csv.html#csv.QUOTE_MINIMAL
        na_value: Value to write that represents empty or missing data.
            See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
            for the na_values supported by default in pandas.read_csv
        nrows: Optional, Number of rows to write. If None, then write all rows.
        allow_dataset_refresh: If True, allows running a job to refresh dataset to make streamable.
            Otherwise a RuntimeError will be thrown if the dataset is unstreamable.
        buffer_size: Number of records to store in memory before writing to disk
        overwrite: if True and export_file_name already exists, overwrite the file.
            Otherwise throw an error
        encoding: The encoding to use in the written file.
            See https://docs.python.org/3/library/functions.html#open

    Returns:
        The total number of records written

    Raises:
        FileExistsError: if the csv file to which the dataset is to be streamed exists
            and `overwrite` is False
        RuntimeError: if `dataset` is not streamable and `allow_dataset_refresh` is False
        ValueError: if `columns` or `flatten_columns` contain columns that are not
            present in `dataset`, or if column renaming would yield duplicate column names
    """
    LOGGER.info(
        f"Streaming records to csv file {export_file_path} from dataset {dataset.name} "
        f"(id={dataset.resource_id})."
    )

    if os.path.exists(export_file_path):
        if not overwrite:
            message = (
                f"CSV file {export_file_path} already exists. "
                f"(Set 'overwrite' flag to True if you wish to overwrite)"
            )
            LOGGER.error(message)
            raise FileExistsError(message)
        else:
            LOGGER.warning(f"CSV file {export_file_path} already exists and will be overwritten")

    if csv_delimiter == flatten_delimiter:
        message = (
            f"The CSV delimiter '{csv_delimiter}' cannot be identical to the list "
            f"flattening delimiter '{flatten_delimiter}'"
        )
        LOGGER.error(message)
        raise ValueError(message)

    attribute_names = [attr.name for attr in dataset.attributes]

    # check that specified columns exist
    if columns is not None:
        common._check_columns_subset(
            input_list=columns, reference_list=attribute_names, raise_error=True
        )

    if not dataset.status().is_streamable:
        if allow_dataset_refresh:
            LOGGER.info(f"Refreshing dataset {dataset.name} to make streamable.")
            op = dataset.refresh()
            operation.enforce_success(op)
        else:
            message = (
                f"Dataset {dataset.name} is not streamable. Refresh it first, or "
                f"run with allow_dataset_refresh=True"
            )
            LOGGER.error(message)
            raise RuntimeError(message)

    func = partial(common._flatten_list, delimiter=flatten_delimiter, force=True)

    full_column_name_dict = common._get_column_mapping_dict(
        dataset_attribute_names=attribute_names, column_name_dict=column_name_dict, columns=columns
    )

    # Open CSV file and use newline='' as recommended by
    # https://docs.python.org/3/library/csv.html#csv.writer
    with open(export_file_path, "w", newline="", encoding=encoding) as csv_file:
        csv_writer = csv.writer(
            csv_file, delimiter=csv_delimiter, quotechar=quote_character, quoting=quoting
        )
        buffer = []
        header = None
        # Set record number to -1 in case the dataset streamed has no records
        record_number = -1

        for record_number, record in enumerate(
            common._yield_records(dataset, func=func, columns=columns)
        ):
            # Obtain and write the header information only on the first pass
            if header is None:
                header = full_column_name_dict.values()
                csv_writer.writerow(header)

            # Replace empty values with a specific null value
            # This also allows nulls to be treated differently from empty strings
            record = [
                na_value if record[k] is None else record[k] for k in full_column_name_dict.keys()
            ]
            buffer.append(record)

            at_max_buffer = buffer_size is not None and (len(buffer) >= buffer_size)
            at_max_rows = nrows is not None and record_number >= nrows - 1
            if at_max_buffer or at_max_rows:
                csv_writer.writerows(buffer)
                LOGGER.debug(f"Written dataset {dataset.name} up to record {record_number+1}")
                buffer = []
                if at_max_rows:
                    break

        # Write anything remaining
        # This will occur whenever the buffer is non-zero and the number of records
        # is not exactly divisible by the buffer number
        # For example, writing a dataset with 1100 records using a buffer size of 500
        # will write in 3 chunks: 2 x 500 above and the remaining 100 handled here
        if len(buffer) != 0:
            LOGGER.debug(f"Written dataset {dataset.name} up to record {record_number + 1}")
            csv_writer.writerows(buffer)

        if record_number == -1:
            # If record number is -1 then no records were streamed, possibly because the dataset
            # has no records. We therefore want to simply save the headers
            if columns is not None:
                csv_writer.writerow(columns)
            else:
                csv_writer.writerow(attribute_names)

    records_written = record_number + 1

    LOGGER.info(
        f"Wrote {records_written} from dataset {dataset.name} (id={dataset.resource_id}) "
        f"to {export_file_path}"
    )

    return records_written


def from_taxonomy(
    project: Project,
    export_file_path: Union[Path, str],
    *,
    csv_delimiter: str = ",",
    flatten_delimiter: str = "|",
    quote_character: str = '"',
    quoting: int = csv.QUOTE_MINIMAL,
    overwrite: bool = False,
    encoding: str = "utf-8",
) -> int:
    """
    Export a Tamr taxonomy to a csv file. Records are streamed to disk and written according to a
    given buffer size.

    Args:
        project: Tamr Project object
        export_file_path: Path to the csv file where the dataset will be saved
        csv_delimiter: Delimiter of the csv file
        flatten_delimiter: Flatten list types to strings by concatenating with this delimiter
        quote_character: Character used to escape value for csv delimiter when it appears in the
            value.
        quoting: The escape strategy to use according to the Python csv writer.
            See https://docs.python.org/2/library/csv.html#csv.QUOTE_MINIMAL
        overwrite: if True and export_file_name already exists, overwrite the file.
            Otherwise throw an error
        encoding: The encoding to use in the written file.
            See https://docs.python.org/3/library/functions.html#open

    Returns:
        The total number of records written

    Raises:
        FileExistsError: if `export_file_path` exists and `overwrite` is set to False
        IOError: if the specified filepath does not exist or cannot be accessed
        RuntimeError: if the classification project is not yet associated with a taxonomy or
                      taxonomy cannot be written to a csv file
        TypeError: if the project type is not classification
        ValueError: if `columns` and `flatten_columns` are identical values
    """
    LOGGER.info(
        f"Streaming taxonomy to csv file {export_file_path} from project {project.name} "
        f"(project id={project.resource_id})."
    )

    try:
        project = project.as_categorization()
    except TypeError:
        not_categorization_error = f"Project {project.name} is not a classification project."
        LOGGER.error(not_categorization_error)
        raise TypeError(not_categorization_error)

    if os.path.exists(export_file_path):
        if not overwrite:
            message = (
                f"CSV file {export_file_path} already exists. "
                f"(Set 'overwrite' flag to True if you wish to overwrite)"
            )
            LOGGER.error(message)
            raise FileExistsError(message)
        else:
            LOGGER.warning(f"CSV file {export_file_path} already exists and will be overwritten")

    if csv_delimiter == flatten_delimiter:
        message = (
            f"The CSV delimiter '{csv_delimiter}' cannot be identical to the list "
            f"flattening delimiter '{flatten_delimiter}'"
        )
        LOGGER.error(message)
        raise ValueError(message)

    try:
        taxonomy = project.as_categorization().taxonomy()
    except requests.exceptions.RequestException:
        no_taxonomy_error = f"Project {project.name} is not associated with any taxonomy yet."
        LOGGER.error(no_taxonomy_error)
        raise RuntimeError(no_taxonomy_error)

    # obtain categories and store in a list
    categories = taxonomy.categories()
    taxonomy_list = []
    for category in categories:
        taxonomy_list.append(category.path)

    # sort the categories
    taxonomy_list.sort()

    # Open CSV file and use newline='' as recommended by
    # https://docs.python.org/3/library/csv.html#csv.writer

    try:
        f = open(export_file_path, "w", newline="", encoding=encoding)
    except (FileNotFoundError, IOError, PermissionError):
        cannot_open_error = f"File path {export_file_path} could not be opened for writing."
        LOGGER.error(cannot_open_error)
        raise IOError(cannot_open_error)
    else:
        try:
            csv_writer = csv.writer(
                f, delimiter=csv_delimiter, quotechar=quote_character, quoting=quoting
            )
            csv_writer.writerows(taxonomy_list)
        except csv.Error as e:
            general_error = (
                "Encountered an error while writing taxonomy categories to "
                f"{export_file_path}: {e}"
            )
            f.close()
            LOGGER.error(general_error)
            raise RuntimeError(general_error)
        finally:
            f.close()

    records_written = len(taxonomy_list)
    LOGGER.info(
        f"Wrote {records_written} categories from {project.name} taxonomy (project id"
        f"={project.resource_id}) to {export_file_path}"
    )
    return records_written



import logging
import os
from datetime import datetime, timedelta
from typing import Optional

from tamr_toolbox.enrichment.address_mapping import AddressValidationMapping
from tamr_toolbox.models.data_type import JsonDict

# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke docs` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    import googlemaps


LOGGER = logging.getLogger(__name__)
_ADDRESSVALIDATION_BASE_URL = "https://addressvalidation.googleapis.com"


def get_maps_client(googlemaps_api_key: str) -> "googlemaps.Client":
    """Get GoogleMaps client.

    Args:
        googlemaps_api_key: API key for the Google Maps address validation API

    Returns:
        API client linekd to specified key
    """
    import googlemaps

    return googlemaps.Client(key=googlemaps_api_key)


def validate(
    *,
    address_to_validate: str,
    client: "googlemaps.Client",
    locality: Optional[str] = None,
    region_code: Optional[str] = None,
    enable_usps_cass: bool = False,
) -> Optional[AddressValidationMapping]:
    """Validate an address using google's address validation API.

    Args:
        address_to_validate:  address to validate
        client: client for Google Maps API
        region_code: region to use for validation, optional
        locality: locality to use for validation, optional
        enable_usps_cass: whether to use USPS validation

    Returns:
        toolbox address validation mapping, or None if API call fails

    Raises:
        RuntimeError: if API key is invalid
    """
    params = {"address": {"addressLines": [address_to_validate]}}

    if region_code:
        params["address"]["regionCode"] = region_code

    if locality:
        params["address"]["locality"] = locality

    if enable_usps_cass:
        params["enableUspsCass"] = True

    json_resp: JsonDict = client._request(
        "/v1:validateAddress",
        {},  # No GET params
        base_url=_ADDRESSVALIDATION_BASE_URL,
        extract_body=lambda x: x.json(),
        post_json=params,
    )

    if "api key not valid" in json_resp.get("error", dict()).get("message", "").lower():
        message = "Invalid API key supplied to Google Maps address validation"
        LOGGER.error(message)
        raise RuntimeError(message)

    if not json_resp.get("result"):
        message = f"Got no result for {address_to_validate}: API returned {json_resp}."
        LOGGER.error(message)
        raise RuntimeError(message)

    # Parse the response to extract the desired fields
    json_resp = json_resp["result"]

    # Get USPS address components
    usps_address: JsonDict = json_resp.get("uspsData", dict()).get("standardizedAddress", dict())
    usps_zipcode = usps_address.get("zipCode")
    if usps_zipcode and usps_address.get("zipCodeExtension"):
        usps_zipcode += "-" + usps_address.get("zipCodeExtension")

    # Get postal address components
    postal_address = json_resp.get("address", dict()).get("postalAddress", dict())

    return AddressValidationMapping(
        input_address=address_to_validate,
        validated_formatted_address=json_resp["address"]["formattedAddress"],
        expiration=str(datetime.now() + timedelta(days=30)),
        region_code=postal_address.get("regionCode"),
        postal_code=postal_address.get("postalCode"),
        admin_area=postal_address.get("administrativeArea"),
        locality=postal_address.get("locality"),
        address_lines=postal_address.get("addressLines"),
        usps_first_address_line=usps_address.get("firstAddressLine"),
        usps_city_state_zip_line=usps_address.get("cityStateZipAddressLine"),
        usps_city=usps_address.get("city"),
        usps_state=usps_address.get("state"),
        usps_zip_code=usps_zipcode,
        latitude=json_resp.get("geocode", dict()).get("location", dict()).get("latitude"),
        longitude=json_resp.get("geocode", dict()).get("location", dict()).get("longitude"),
        place_id=json_resp.get("geocode", dict()).get("placeId"),
        input_granularity=json_resp.get("verdict", dict()).get(
            "inputGranularity", "GRANULARITY_UNSPECIFIED"
        ),
        validation_granularity=json_resp.get("verdict", dict()).get(
            "validationGranularity", "GRANULARITY_UNSPECIFIED"
        ),
        geocode_granularity=json_resp.get("verdict", dict()).get(
            "geocodeGranularity", "GRANULARITY_UNSPECIFIED"
        ),
        has_inferred=json_resp.get("verdict", dict()).get("hasInferredComponents", False),
        has_unconfirmed=json_resp.get("verdict", dict()).get("hasUnconfirmedComponents", False),
        has_replaced=json_resp.get("verdict", dict()).get("hasReplacedComponents", False),
        address_complete=json_resp.get("verdict", dict()).get("addressComplete", False),
    )



"""
Additional functions to manipulate the profile of the dataset.
"""

from tamr_unify_client.dataset.resource import Dataset, DatasetProfile
from requests import HTTPError
import logging

LOGGER = logging.getLogger(__name__)


def get_profile(dataset: Dataset, allow_create_or_refresh: bool = False) -> DatasetProfile:
    """
    Returns a dataset profile object.  Optionally can refresh or create profile if missing or
    out-of-date.
    Args:
        dataset: Tamr dataset object
        allow_create_or_refresh: optional bool to allow creation/refreshing of profile info
    Returns:
        DatasetProfile object
        Warning if profile information is out of date and allow_create_or_refresh is False
    Raises:
        RuntimeError: if profile has not been created and allow_create_or_refresh is False
    """
    try:
        profile = dataset.profile()
    except HTTPError as e:
        # profile not yet created or in inconsistent state: 409 or 400 return respectively
        if e.response.status_code in [409, 400]:
            if allow_create_or_refresh:
                LOGGER.info(f"creating profile information for dataset: {dataset.name}")
                dataset.create_profile()
                profile = dataset.profile()
            else:
                raise RuntimeError(
                    f"Profile information for {dataset.name} does not exist. If you would like "
                    f"to create one, set allow_create_or_refresh to True."
                )
        # other status codes: re-raise error
        else:
            raise e

    if not profile.is_up_to_date:
        if allow_create_or_refresh:
            LOGGER.info(f"refreshing profile information for dataset: {dataset.name}")
            profile.refresh()
            # retrieve again for up-to-date information
            profile = dataset.profile()
        else:
            # Return profile information with a staleness warning:
            print(
                f"WARNING: Profile information for {dataset.name} is out-of-date "
                f"and allow_create_or_refresh is False. If you would like an up-to-date "
                f"profile, rerun with allow_create_or_refresh set to True."
            )

    return profile



import logging
from typing import List, Set, Tuple

from dataclasses import dataclass
import networkx as nx

from tamr_unify_client import Client
from tamr_unify_client.project.resource import Project
from tamr_toolbox.models.project_type import ProjectType
from typing import Dict, Any

LOGGER = logging.getLogger(__name__)


@dataclass()
class Graph:
    """
    A dataclass for holding a set of Tamr project dependencies (edges),
    and the generated graph from them.
    """

    edges: set
    directed_graph: nx.DiGraph


def _get_upstream_projects(project: Project, *, all_projects: List[Project]) -> List[Project]:
    """
    get projects immediately upstream of a given project
    Args:
        project: the project to check
        all_projects: a list of all Projects on the Tamr Core instance

    Returns:
        A list of project names upstream of the project
    """
    client = project.client
    # find upstream datasets - if GR project just get input datasets
    if ProjectType[project.type] == ProjectType.GOLDEN_RECORDS:
        upstream_datasets = [x for x in project.input_datasets().stream()]
    # else find the upstream datasets of the UD (not input datasets to capture datasets used in Tx)
    else:
        unified_dataset_id = project.unified_dataset().relative_id
        unified_dataset = client.datasets.by_relative_id(unified_dataset_id)
        upstream_datasets = unified_dataset.upstream_datasets()

    upstream_project_names = []
    # walk through upstream datasets
    for upstream_result in upstream_datasets:
        # get the upstream object as a dataset
        upstream_dataset = client.datasets.by_resource_id(upstream_result.resource_id)
        # see if it is the output of a project and if so add to the list
        upstream_dataset_projects = set(
            x.project_name for x in upstream_dataset.usage().usage.output_from_project_steps
        )
        upstream_project_names.extend([x for x in upstream_dataset_projects])

    # we have all projects in all_projects variables so only return those whose
    # name is in upstream_project_names
    return [x for x in all_projects if x.name in upstream_project_names]


def _build_edges(
    project: Project, client: Client, *, edges: set = None, all_projects: List[Project]
) -> Set[Tuple[str, str]]:
    """
    builds a set of tuples of all edges of format (source, target)
    Args:
        project: the project to get edges for
        client: tamr client
        edges: set of tuples (source, target)
        all_projects: a list of all Projects on the Tamr Core instance

    Returns:

    """
    upstream_projects = _get_upstream_projects(project, all_projects=all_projects)

    if edges is None:
        edges = set()

    for upstream_project in upstream_projects:
        # add the edge for this upstream dataset
        edges.add((upstream_project.name, project.name))
        # if we've already walked backward for this upstream dataset don't keep walking
        # you know we've walked it if it shows up as the target in an edge (hence the [1] index)
        if any(upstream_project.name == x[1] for x in edges):
            LOGGER.debug(
                f"skipping dataset {upstream_project.name} since it is already"
                f" in edges as target: {[ x for x in edges if x[1] ==upstream_project.name]}"
            )
            continue
        else:
            # and then go to it and get it's upstream datasets
            further_upstream_edges = _build_edges(
                upstream_project, client, edges=set(edges), all_projects=all_projects
            )
            # print(f"adding further upstream edges {further_upstream_edges}")
            edges = edges.union(further_upstream_edges)

    return edges


def from_edges(edges: Set[tuple]) -> Graph:
    """
    Directly build a graph from a list of edges - tuples of format (source, target) dependencies

    Args:
        edges: List of edges in tuple format

    Returns:
        Graph object
    """
    graph = nx.DiGraph()
    graph.add_edges_from(edges)
    return Graph(edges=edges, directed_graph=graph)


def from_project_list(projects: List[Project], client: Client) -> Graph:
    """
    Creates a graph from a list of datasets
    Args:
        projects: list of Tamr dataset objects
        client: tamr client

    Returns:
        A Graph object built from the dependencies of the datasets passed
    """
    # start with empty set
    graph_edges = set()
    # save all projects to list so we don't have to hit the API every time
    all_projects = [x for x in client.projects.stream()]
    # for each dataset get the edges and take union
    for project in projects:
        graph_edges = set(
            graph_edges.union(_build_edges(project, client, all_projects=all_projects))
        )

    graph = nx.DiGraph()
    graph.add_edges_from(graph_edges)
    return Graph(edges=graph_edges, directed_graph=graph)


def get_source_nodes(graph: Graph) -> List[str]:
    """
    Gives all source nodes in a graph

    Args:
        graph: Graph for which to find source nodes

    Returns:
        List of node names
    """
    di_graph = graph.directed_graph
    source_nodes = []
    for n in di_graph.nodes():
        if not [x for x in di_graph.predecessors(n)]:
            source_nodes.append(n)
    return source_nodes


def get_end_nodes(graph: Graph):
    """
    Returns all end nodes in a directed graph

    Args:
        graph: Graph for which to find end nodes

    Returns:
        List of names of all end nodes
    """
    di_graph = graph.directed_graph
    end_nodes = []
    for n in di_graph.nodes():
        if not [x for x in di_graph.successors(n)]:
            end_nodes.append(n)
    return end_nodes


def get_projects_by_tier(graph: Graph) -> Dict[int, Any]:
    """
    Find the different projects at each tier

    Args:
        graph: the Graph for which to generate the tiers

    Returns:
        A json dict who's structure is {'tier': [paths_at_that_tier], ...}
        e.g. {1: ['SM_project_1', 'Classification_project_1'], 2: ['Mastering_project'],
        3: ['Golden_records_project']}
    """
    source_nodes = get_source_nodes(graph)

    path_tier_dict = {0: []}
    for n in graph.directed_graph.nodes:
        # just add source nodes to tier 0
        if n in source_nodes:
            path_tier_dict[0].append(n)
            continue
        # since all dependent projects must be run first
        # the tier is the maximum length of all simple paths to this node
        # so get the max length from all source nodes
        tier = 0
        for s in source_nodes:
            current_paths = [x for x in nx.all_simple_paths(graph.directed_graph, s, n)]
            # skip source nodes that don't link to this project
            if not current_paths:
                continue
            max_path_length = len(max(current_paths, key=lambda x: len(x)))
            if max_path_length > tier:
                tier = max_path_length

        # decrement tier to count from 0
        tier = tier - 1
        # check if tier exists in the dict, if so add to it, else create entry
        if tier in path_tier_dict:
            path_tier_dict[tier].append(n)
        else:
            path_tier_dict[tier] = [n]

    return {k: set(v) for k, v in path_tier_dict.items()}


def get_all_downstream_nodes(graph: Graph, node: str) -> Set[str]:
    """
    Get all nodes downstream of this one (i.e. they have a path from this node to them)

    Args:
        graph: the graph to use
        node: the node to check

    Returns:
        A list of downstream node names
    """
    downstream_paths = []
    diGraph = graph.directed_graph
    for n in diGraph.nodes():
        all_paths = [x for x in nx.all_simple_paths(diGraph, node, n)]
        if all_paths:
            for path in all_paths:
                downstream_paths.extend(path)

    # the above generates a list of things like
    # [projectA, projectB, projectA, projectC] if projectA goes to both project_B and project_C
    # so make a set, then remove the node itself
    downstream_nodes = {x for x in set(downstream_paths) if x != node}
    return downstream_nodes


def get_successors(graph: Graph, node: str) -> Set[str]:
    """
    Get all successor nodes to the current node

    Args:
        graph: the graph to use
        node: the node to check

    Returns:
        A set of nodes that are successors to the current node
    """
    di_graph = graph.directed_graph
    return set(x for x in di_graph.successors(node))


def get_predecessors(graph: Graph, node: str) -> Set[str]:
    """
    Get all predecessor nodes to the current node

    Args:
        graph: the graph to use
        node: the node to check

    Returns:
        A set of nodes that are predecessors to the current node
    """
    di_graph = graph.directed_graph
    return set(x for x in di_graph.predecessors(node))


def add_edges(graph: Graph, edges: Set[tuple]) -> Graph:
    """
    Takes an existing graph and creates a new one with the new edge

    Args:
        graph: the graph to start with
        edges: the edges to add

    Returns:
        A copy of initial graph with new edge
    """

    old_edges = set(graph.directed_graph.edges)
    new_edges = old_edges.union(edges)
    return from_edges(new_edges)



"""Tasks related to Tamr Mastering projects"""
from . import jobs
from .._common import attribute_configuration, schema, transformations

__all__ = ["jobs", "attribute_configuration", "schema", "transformations"]



"""Tasks related to moving data in or out of Tamr"""
from . import dataframe
from . import df_connect
from . import common
from . import file_system_type
from . import csv

__all__ = ["common", "csv", "dataframe", "df_connect", "file_system_type"]



import logging
from tamr_unify_client.dataset.resource import Dataset
from tamr_unify_client.project.resource import Project
from tamr_unify_client.client import Client
from typing import List, Optional, Dict
from collections import OrderedDict
from tamr_toolbox.models.project_type import ProjectType
import re

LOGGER = logging.getLogger(__name__)


def datasets(dataset: Dataset, *, include_dependencies_by_name: bool = False) -> List[Dataset]:
    """Returns a dataset's downstream datasets.

    Args:
        dataset: The target dataset.
        include_dependencies_by_name: Whether to include datasets based on name similarity.
            No dependencies will be found by name if the `dataset` is not an unified dataset
            either based on backened pipeline (if project still exists) or based on
            regex (dataset name has suffix 'unified_dataset').

    Returns:
        List of Dataset objects ordered by number of its downstream dependencies.
            Note that there can be bidirectional dependency so datasets with same number of
            dependencies can depend on each other.
    """
    downstream_datasets = _find_downstream_datasets(
        dataset.name,
        include_dependencies_by_name=include_dependencies_by_name,
        client=dataset.client,
    )
    # Remove the target dataset from the list
    downstream_datasets.pop(dataset.name)
    # Extract the list of the Dataset objects
    downstream_datasets = list(downstream_datasets.values())
    # sort list of downstream datasets by number of dependencies
    # because parent dataset always has greater number of dependencies than its child datasets
    # note that there can be bidirectional dependency so datasets with same number of
    # dependencies can depend on each other
    LOGGER.info("Sorting the downstream datasets by number of dependencies")
    sorted_datasets = sorted(downstream_datasets, key=lambda x: len(x.usage().dependencies))
    return sorted_datasets


def projects(dataset: Dataset, *, include_dependencies_by_name: bool = False) -> List[Project]:
    """Return list of downstream project_list for a dataset.

    Args:
        dataset: The target dataset.
        include_dependencies_by_name: Whether to include datasets based on name similarity.
            No dependencies will be found by name if the `dataset` is not an unified dataset
            either based on backened pipeline (if project still exists) or based on
            regex (dataset name has suffix 'unified_dataset').

    Returns:
        List of downstream project_list in order,
            including the project the target `dataset` is part of.

    """
    downstream_datasets = datasets(
        dataset, include_dependencies_by_name=include_dependencies_by_name
    )
    project_ids = []
    # downstream dataset list is ordered by number of dependency ascending
    for dependency in downstream_datasets:
        project_ids.extend(_find_associated_projects(dependency))
    # add project_list of the target dataset
    project_ids.extend(_find_associated_projects(dataset))

    # use Ordered Dict to keep the order (set doesn't keep the order)
    project_ids = OrderedDict().fromkeys(project_ids)
    LOGGER.debug(f"Downstream project ids found: {project_ids}")

    # Need to reverse to get the most upstream project first
    project_list = [
        dataset.client.projects.by_resource_id(i) for i in reversed(project_ids.keys())
    ]
    return project_list


def _find_downstream_datasets(
    dataset_name: str,
    *,
    client: Client,
    include_dependencies_by_name: bool = False,
    all_tamr_datasets: Optional[Dict[str, Dataset]] = None,
    all_unified_datasets: Optional[List[str]] = None,
    downstream_datasets: Optional[Dict[str, Dataset]] = None,
) -> Dict[str, Dataset]:
    """Returns a dataset's downstream datasets.

    Args:
        dataset_name: Name of the target dataset.
        client: Tamr client
        include_dependencies_by_name: Whether to include datasets based on name similarity.
        all_tamr_datasets: A dictionary of all datasets existing in Tamr by name.
        all_unified_datasets: A List of current unified datasets in Tamr.
        downstream_datasets: A dictionary of downstream datasets by name.
            This argument is not to be set by users, and only used internally.

    Returns:
        List of Dataset objects.

    """
    # initialize optional variables
    if not all_tamr_datasets:
        all_tamr_datasets = {dataset.name: dataset for dataset in client.datasets}
    if not all_unified_datasets:
        all_unified_datasets = _get_all_unified_datasets(client)
    if not downstream_datasets:
        downstream_datasets = {}

    # Initialize a set of datasets need to check further downstream
    # (if include dependencies by name)
    datasets_check = set()
    LOGGER.debug(f"Start checking dependencies of {dataset_name}")
    dataset = all_tamr_datasets[dataset_name]
    dataset_dependencies = dataset.usage().dependencies
    LOGGER.debug(
        f"Dataset {dataset.name} has {len(dataset_dependencies)} "
        f"datasets with dependency suggested by pipeline"
    )
    for d in dataset_dependencies:
        if include_dependencies_by_name and _is_unified_dataset(
            d.dataset_name, all_unified_datasets=all_unified_datasets
        ):
            # if a downstream dataset is unified dataset and want to include dependencies by name,
            # need to check it further downstream
            datasets_check.add(d.dataset_name)
        elif d.dataset_name not in downstream_datasets.keys():
            downstream_datasets[d.dataset_name] = d.dataset()

    # get dependencies suggested by name matching
    # if the target dataset is or used to be unified dataset
    # and whose prefix is the target dataset's name if the target is unified dataset
    if include_dependencies_by_name and _is_unified_dataset(
        dataset_name, all_unified_datasets=all_unified_datasets
    ):
        dependencies_suggested_by_name = set(
            [d for d in all_tamr_datasets.keys() if d.startswith(dataset_name)]
        )
        LOGGER.debug(
            f"Dataset {dataset_name} has {len(dependencies_suggested_by_name)} "
            f"datasets with dependency suggested by name"
        )
        datasets_check = datasets_check.union(dependencies_suggested_by_name)
        # Filter out datasets that have already been checked and the target dataset
        datasets_check = datasets_check.difference(downstream_datasets.keys()).difference(
            {dataset.name}
        )

    # add the target dataset to list of downstream datasets to avoid checking later
    downstream_datasets[dataset_name] = dataset

    # check further downstream for either dependencies suggested by name
    # or dependencies that used to or is unified dataset
    # note: dictionary `downstream_datasets` is passed by reference
    for d in datasets_check:
        _find_downstream_datasets(
            d,
            include_dependencies_by_name=include_dependencies_by_name,
            client=client,
            all_tamr_datasets=all_tamr_datasets,
            all_unified_datasets=all_unified_datasets,
            downstream_datasets=downstream_datasets,
        )

    return downstream_datasets


def _find_associated_projects(dataset: Dataset) -> List[str]:
    """Return list of project_list that the dataset is part of.

    Args:
        dataset: The target dataset.

    Returns:
        List of Project IDs that the target `dataset` is part of.

    """
    project_list = set(
        [step.project().resource_id for step in dataset.usage().usage.input_to_project_steps]
    )
    return list(project_list)


def _is_unified_dataset(
    dataset_name: str,
    *,
    check_regex: bool = True,
    regex=".*unified_dataset$",
    all_unified_datasets: List[str],
) -> bool:
    """Check if a dataset is or used to be unified dataset.

    Args:
        dataset_name: Name of the target dataset.
        check_regex: A Boolean to determine whether to check for regex matching if a dataset is
            not part of the pipeline (if its project has been deleted.) Default to True.
        regex: A string of regex to check. Default to '.*unified_dataset$' which is Tamr's current
            default unified dataset naming.
        all_unified_datasets: List of unified dataset names.

    Returns:
        A Boolean, if True, the target `dataset` is or used to be unified dataset.

    """
    # check project pipeline (only works if project has not been deleted)
    if dataset_name in all_unified_datasets:
        return True
    # check regex
    if check_regex:
        return re.search(regex, dataset_name) is not None
    return False


def _get_all_unified_datasets(client: Client) -> List[str]:
    """

    Args:
        client: Tamr client

    Returns:
        List of unified dataset names.

    """
    # Use Project Spec to minimize number of API calls
    unified_dataset_names = [
        project.spec().to_dict()["unifiedDatasetName"]
        for project in client.projects
        # GOLDEN_RECORDS project does not have unified dataset
        if ProjectType[project.type]
        in [
            ProjectType.SCHEMA_MAPPING_RECOMMENDATIONS,
            ProjectType.DEDUP,
            ProjectType.CATEGORIZATION,
        ]
    ]
    return unified_dataset_names



import json
import copy
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import Dict, List
import logging

from tamr_unify_client import Client

from tamr_toolbox.workflow.concurrent.Graph import (
    Graph,
    get_projects_by_tier,
    get_all_downstream_nodes,
    get_successors,
    get_predecessors,
)
from tamr_toolbox.workflow.concurrent import PlanNodeStatus
from tamr_toolbox.workflow.concurrent.PlanStatus import PlanStatus, from_planner
from tamr_toolbox.workflow.concurrent.PlanNode import PlanNode, run_next_step, monitor


LOGGER = logging.getLogger(__name__)


@dataclass_json
@dataclass
class Planner:
    """
    A dataclass to hold the plan, the starting tier, and the mode of execution.
    The plan is a json dict where each key is a project name and the value is a PlanNode object

    The starting tier is the tier at which to start execution. All jobs at lower tiers are
     marked as skippable.

    The graph is the graph that contains the backing project dependencies.
    """

    plan: Dict[str, PlanNode]
    starting_tier: int
    graph: Graph


def from_graph(
    graph: Graph, *, tamr_client: Client, starting_tier: int = 0, train=False
) -> Planner:
    """
    Creates a Planner class from a Graph. The plan object is a json dict specifying how
    the plan can be executed and its status.

    Args:
        graph: the dataset dependency graph to use to create the planner
        tamr_client: the tamr client object associated with the instance for which
            to create the plan
        starting_tier: the tier at which to start executing the plan, every job at lower
            tiers is skipped and marked as skippable
        train: global config for whether or not to 'apply feedback'/train the model in
            the workflows

    Returns:
        Planner instance
    """

    # graphs don't store tamr project objects themselves just the names so need to build lookup
    tamr_projects = {x.name: x for x in tamr_client.projects.stream()}

    # start with the project tier dict from the graph
    tier_graph = get_projects_by_tier(graph)

    # start building the plan
    plan = {}
    for tier, project_list in tier_graph.items():
        for num, project_name in enumerate(project_list):
            # mark things as skippable if the tier is less than the starting tier
            if tier < starting_tier:
                status = PlanNodeStatus.PlanNodeStatus.SKIPPABLE
            elif tier == starting_tier:
                status = PlanNodeStatus.PlanNodeStatus.RUNNABLE
            else:
                status = PlanNodeStatus.PlanNodeStatus.PLANNED
            plan[project_name] = PlanNode(
                priority=(100 * tier) + num,
                status=status,
                name=project_name,
                current_op=None,
                operations=None,
                project=tamr_projects[project_name],
                train=train,
            )

    return Planner(plan=plan, starting_tier=starting_tier, graph=graph)


def update_plan(planner: Planner, *, plan_node: PlanNode) -> Planner:
    """
    Create an new planner object with updated status from an updated PlanNode object
    Args:
        planner: the original planner
        plan_node: an updated set of plan nodes

    Returns:
        a copy of the original planner object with an updated status
    """
    # first update the planner to have the new node contents
    original_plan = planner.plan
    updated_plan = copy.deepcopy(original_plan)
    LOGGER.info(
        f"Updating plan with changed project status: {plan_node.name} "
        f"status changed to {plan_node.status}"
    )
    plan_node_name = plan_node.name
    updated_plan[plan_node_name] = plan_node

    # now get the status of the updated node
    node_status = plan_node.status
    # now find downstream affects
    downstream_nodes = get_all_downstream_nodes(planner.graph, plan_node_name)
    # if status == failed then easy to update them all to blocked
    if (
        node_status == PlanNodeStatus.PlanNodeStatus.FAILED
        or node_status == PlanNodeStatus.PlanNodeStatus.CANCELED
    ):
        for node in downstream_nodes:
            updated_plan[node].status = PlanNodeStatus.PlanNodeStatus.BLOCKED
    # else if update is skippable or successful
    # then need to see if we can mark successor nodes as runnable
    elif (
        node_status == PlanNodeStatus.PlanNodeStatus.SUCCEEDED
        or node_status == PlanNodeStatus.PlanNodeStatus.SKIPPABLE
    ):
        # first get immediate downstream nodes
        successor_nodes = get_successors(planner.graph, plan_node_name)
        # for each of these get the predecessors and if all predecessors are now
        # succeeded/skippable mark as runnable
        for successor in successor_nodes:
            predecessor_nodes = get_predecessors(planner.graph, successor)
            if all(
                updated_plan[x].status == PlanNodeStatus.PlanNodeStatus.SUCCEEDED
                or updated_plan[x].status == PlanNodeStatus.PlanNodeStatus.SKIPPABLE
                for x in predecessor_nodes
            ):
                updated_plan[successor].status = PlanNodeStatus.PlanNodeStatus.RUNNABLE

    # no other status should change things
    # so do nothing else

    return Planner(plan=updated_plan, graph=planner.graph, starting_tier=planner.starting_tier)


def to_json(planner: Planner) -> List[Dict]:
    """
    Convert planner to a json dict
    Args:
        planner: the planner to convert

    Returns:
        representation of a planner in json format
    """
    return [
        {"name": v.name, "status": v.status, "priority": v.priority}
        for k, v in planner.plan.items()
    ]


def execute(
    planner: Planner,
    tamr: Client,
    *,
    concurrency_level: int = 2,
    save_state: bool = False,
    polling_interval: int = 30,
) -> Planner:
    """
    Executes the plan

    Args:
        planner: The planner object whose plan will be executed
        tamr: the tamr client to use
        concurrency_level: the number of concurrent jobs to run at once
        save_state: whether or not to save the plan state to json after each update
        polling_interval: the amount of time in seconds to wait between polling

    Returns:
        the planner object after execution
    """
    # get the plan and sort by priority
    plan = planner.plan
    sorted_jobs = [v for k, v in sorted(plan.items(), key=lambda x: x[1].priority)]
    # assume you could be given a partially executed plan
    # so create runnable, pending, and running queues
    runnable_nodes = [x for x in sorted_jobs if x.status == PlanNodeStatus.PlanNodeStatus.RUNNABLE]
    running_nodes = [x for x in sorted_jobs if x.status == PlanNodeStatus.PlanNodeStatus.RUNNING]
    pending_nodes = [
        x for x in sorted_jobs if x.status == PlanNodeStatus.PlanNodeStatus.PENDING_NEXT_STEP
    ]

    # check status and run if runnable or planned
    plan_status = from_planner(planner)
    if plan_status == PlanStatus.PLANNED or plan_status == PlanStatus.RUNNING:
        LOGGER.info(
            "projects with currently running jobs: "
            f"{','.join([x.name for x in running_nodes]) or 'None'}"
        )
        # make sure there are fewer jobs running than concurrency specified
        num_to_submit = concurrency_level - len(running_nodes)
        LOGGER.info(f"Have room in queue for {num_to_submit} jobs:")

        # slice runnable jobs to get the ones to submit
        # this line is for type hinting
        nodes_to_submit: List[PlanNode] = []

        # all pending nodes need their next step to be triggered so there is only room
        # in the queue for new nodes depending on how many pending nodes there are
        num_to_add = num_to_submit - len(pending_nodes)
        nodes_to_submit = [x for x in pending_nodes]
        if num_to_add > 0:
            nodes_to_submit.extend([x for x in runnable_nodes[0:num_to_add]])

        LOGGER.info(f"submitting jobs for projects: [{','.join(x.name for x in nodes_to_submit)}]")
        # create the list of nodes to monitor
        # note that this returns a list of plan nodes AND triggers the job
        nodes_to_monitor = [run_next_step(x) for x in nodes_to_submit]
        # extend jobs_to_monitor to include running jobs
        LOGGER.info(
            f"Adding currently running projects to queue: {[x.name for x in running_nodes]}"
        )
        nodes_to_monitor.extend(
            [
                PlanNode(
                    name=x.name,
                    status=x.status,
                    priority=0,
                    current_op=x.current_op,
                    project=x.project,
                    operations=x.operations,
                    steps_to_run=x.steps_to_run,
                    current_step=x.current_step,
                )
                for x in running_nodes
            ]
        )

        # now monitor the nodes
        # this function returns when there is any change in state
        nodes_to_monitor = monitor(nodes_to_monitor, polling_interval=polling_interval)
        LOGGER.info(f"Got updated set of jobs: {nodes_to_monitor}")
        # now update the plan - only monitored jobs should have a change in status
        for node in nodes_to_monitor:
            planner = update_plan(planner, plan_node=node)

        LOGGER.info(f"after recent update plan status is {from_planner(planner)}")

        # if save state then save a copy of the plan:
        if save_state:
            LOGGER.info(f"current Planner state: {json.dumps(to_json(planner))}")
        # planner is updated so now try to execute it again
        planner = execute(
            planner,
            tamr=tamr,
            concurrency_level=concurrency_level,
            save_state=save_state,
            polling_interval=polling_interval,
        )
        return planner

    # if planner isn't runnable then exit
    else:
        LOGGER.info(f"plan status is {plan_status} so returning")
        return planner



"""Tasks related to editing the taxonomy for a tamr categorization project"""
import logging
import os

from tamr_unify_client import Client
import json
import copy

LOGGER = logging.getLogger(__name__)

# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke docs` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    import pandas as pd


def _get_children_nodes(all_categories: list, target_node: str):
    """
    Gets all the children nodes for a given node ID from the full taxonomy

    Args:
        all_categories: List of all categories of the taxonomy returned by categories API
        target_node: Full internal ID of target node to get children for

    Returns: List of children nodes including non-leaf nodes under the given target node
    """
    first_pass_children = [cat for cat in all_categories if cat["parent"] == target_node]
    if len(first_pass_children) == 0:
        # Node is a leaf node:
        children_nodes = []
    else:
        children_nodes = first_pass_children
        # Iterate to get children at subsequent tiers:
        while True:
            new_children_all = []
            for node in first_pass_children:
                new_children = [cat for cat in all_categories if cat["parent"] == node["id"]]
                if len(new_children) > 0:
                    children_nodes.extend(new_children)
                    new_children_all.extend(new_children)
            if len(new_children_all) == 0:
                # No deeper tiers found.
                break
            else:
                # Look for deeper tiers among the nodes found in the last iteration:
                first_pass_children = new_children_all

    return children_nodes


def delete_node(client: Client, project_id: str, path: list, force_delete: bool = False):
    """
    Deletes a node from a taxonomy.

    Args:
        client: Tamr client connected to target instance
        project_id: ID of the categorization project
        path: Full path of the node to be deleted
        force_delete: Optional flag. Default is false. If true, deletes even if there are still
        records assigned to that category. If false, the operation fails with an error.

    Returns: None
    """
    # Get all categories to extract category ID:
    cat_response = client.get(f"projects/{project_id}/taxonomy/categories")
    all_cats = json.loads(cat_response.content)
    target_cat = [cat for cat in all_cats if cat["path"] == path][0]
    target_cat_full_id = target_cat["id"]
    target_cat_id = target_cat_full_id.split("/")[-1]

    # Check if there are any children nodes to the target node:
    children_nodes = _get_children_nodes(all_cats, target_cat_full_id)
    all_children_ids = [cat["id"] for cat in children_nodes]
    all_children_ids.append(target_cat_full_id)

    # Check if there are any records verified to the category to be deleted:
    verified_cats_response = client.get(f"projects/{project_id}/categorizations/labels/records")
    node_records = []
    for line in verified_cats_response.iter_lines():
        record = json.loads(line)
        if "verified" in record:
            if record["verified"]["category"]["categoryId"] in all_children_ids:
                node_records.append(record)

    if len(node_records) == 0:
        LOGGER.info(f"No records found for node {path} or any children nodes. Deleting node.")
        client.delete(f"projects/{project_id}/taxonomy/categories/{target_cat_id}")
    else:
        LOGGER.info(
            f"There are verified records under node {path}. Deleting the node without "
            f"moving these records will result in loss of work. If you wish to proceed "
            f"set force_delete flag to True."
        )
        if force_delete:
            LOGGER.info(f"Force Delete is set to true. Deleting node {path}")
            client.delete(f"projects/{project_id}/taxonomy/categories/{target_cat_id}")
        else:
            delete_node_error = "Force delete is set to false. Exiting without deletion."
            LOGGER.error(delete_node_error)
            raise RuntimeError(delete_node_error)
    return


def rename_node(client: Client, project_id: str, new_name: str, path: list):
    """
    Renames an existing node in the taxonomy.

    Args:
        client: Tamr client connected to target instance
        project_id: ID of the categorization project
        new_name: New name to assign to the leaf node
        path: Full path of the existing leaf node to rename

    Returns: None
    """
    body = {"path": path, "name": new_name}
    # Get all categories to extract category ID:
    cat_response = client.get(f"projects/{project_id}/taxonomy/categories")
    all_cats = json.loads(cat_response.content)
    target_cat = [cat for cat in all_cats if cat["path"] == path][0]
    target_cat_id = target_cat["id"].split("/")[-1]
    LOGGER.info(f"Renaming category id {target_cat_id} in project {project_id} to {new_name}")
    response = client.put(f"projects/{project_id}/taxonomy/categories/{target_cat_id}", json=body)
    # Log an error if the operation failed:
    if not response.ok:
        content = json.loads(response.content)
        rename_node_error = (
            f"Renaming node {target_cat_id} failed with message " f"{content['message']}"
        )
        LOGGER.error(rename_node_error)
        raise RuntimeError(rename_node_error)
    return


def create_node(client: Client, project_id: str, path: list):
    """
    Creates a category with the specified path in the project taxonomy.

    Args:
        client: Tamr client connected to target instance
        project_id: ID of the categorization project
        path: Full path of the new category to be added

    Returns: None
    """
    body = {"name": path[-1], "path": path}
    LOGGER.info(f"Creating new category {path[-1]} in project {project_id}")
    response = client.post(f"projects/{project_id}/taxonomy/categories", json=body)
    # Log an error if the operation failed:
    if not response.ok:
        content = json.loads(response.content)
        create_node_error = f"Creating node {path[-1]} failed with message {content['message']}"
        LOGGER.error(create_node_error)
        raise RuntimeError(create_node_error)
    return


def get_taxonomy_as_dataframe(client: Client, project_id: str) -> "pd.DataFrame":
    """
    Returns the taxonomy for a project given the project ID.

    Args:
        client: Tamr client connected to target instance
        project_id: ID of the categorization project

    Returns:
        Current taxonomy categories as a dataframe

    Raises:
        RuntimeError: if project is not a categorization project or if the taxonomy does not exist
    """
    # This function requires pandas, an optional dependency
    import pandas as pd

    LOGGER.info(f"Retrieving taxonomy for project ID {project_id}")
    response = client.get(f"projects/{project_id}/taxonomy/categories")
    # Check for bad or empty response:
    if not response.ok or (response.text == "[]"):
        empty_response_error = (
            f"There was no taxonomy found for project {project_id}. Check if the project_id is "
            f"correct and if there is an existing taxonomy."
        )
        LOGGER.error(empty_response_error)
        raise RuntimeError(empty_response_error)

    max_depth = 0
    categories = []
    for blob in response.json():
        category = blob["path"]
        max_depth = max(max_depth, len(category))
        categories.append(category)

    tiers = [f"T{i}" for i in range(1, max_depth + 1)]
    taxonomy_dict = {}
    for tier in tiers:
        taxonomy_dict[tier] = []
    for category in categories:
        for tier_index, tier_name in enumerate(tiers):
            node_name = None
            if len(category) > tier_index:
                node_name = category[tier_index]
            taxonomy_dict[tier_name].append(node_name)

    return pd.DataFrame(taxonomy_dict)


def _create_action(record_id: str, category_path: list):
    """
    Internal function to create the actions to POST to the updateRecords API.

    Args:
        record_id: The ID of the record to be labeled.
        category_path: The path of the category to assign to the record.

    Returns: Action in correct format.
    """
    action = {
        "action": "CREATE",
        "recordId": record_id,
        "record": {"verified": {"category": {"path": category_path}}},
    }
    json_row = json.dumps(action).encode("utf-8")
    return json_row


def _batch(iterable, n=500):
    """
    Internal function to batch payload to POST categorizations API
    """
    total_len = len(iterable)
    for ndx in range(0, total_len, n):
        yield iterable[ndx : min(ndx + n, total_len)]


def move_node(
    client: Client,
    project_id: str,
    old_node_path: list,
    new_node_path: list,
    move_verifications: bool = True,
):
    """
    Function to move a node in a taxonomy to a new path. By default, the function will also move
    any verified categorizations under the old node to the new paths.

    Args:
        client: Tamr client connected to the target instance.
        project_id: Project ID of categorization project.
        old_node_path: List of the full path for the node to be moved.
        new_node_path: List of the full path for where the node is to be moved to.
        move_verifications: Optional boolean argument to move verifications to the new path.
        Setting to false may result in loss of work.

    Returns: None
    """
    # First we need to get all the paths of nodes to be moved:
    taxonomy_df = get_taxonomy_as_dataframe(client, project_id)
    all_paths = taxonomy_df.agg(lambda x: list(x.dropna()), axis=1).tolist()
    paths_to_move = [path for path in all_paths if all(x in path for x in old_node_path)]
    paths_to_move.sort()

    # Now we need to generate the new paths:
    new_paths = copy.deepcopy(paths_to_move)
    length_to_replace = len(old_node_path)
    for path in new_paths:
        path[:length_to_replace] = new_node_path

    # Next we want to create all the new paths. This operation is safe if the paths exist:
    for path in new_paths:
        LOGGER.info("Creating the new nodes")
        create_node(client, project_id, path)

    # Then, we want to move any existing verifications from the old to new paths:
    if move_verifications:
        payload = []
        # First get all the verified records:
        verified_cats_response = client.get(
            f"projects/{project_id}/categorizations/labels/records"
        )
        all_records = []
        for line in verified_cats_response.iter_lines():
            all_records.append(json.loads(line))

        # Check if there are any verified responses in the stream:
        verified_recs = [record for record in all_records if "verified" in record]

        # Next, for each node, check if there are any verified responses:
        if len(verified_recs) > 0:
            for i in range(len(paths_to_move)):
                sub_records = [
                    record
                    for record in all_records
                    if record["verified"]["category"]["path"] == paths_to_move[i]
                ]
                if len(sub_records) == 0:
                    continue
                else:
                    # If there are records, create actions to move them:
                    for record in sub_records:
                        action = _create_action(record["recordId"], new_paths[i])
                        payload.append(action)

            # Post verified records in batches:
            LOGGER.info(f"Moving verified records from {old_node_path} to {new_node_path}")
            for batch in _batch(payload):
                client.post(
                    f"projects/{project_id}/categorizations/labels:updateRecords",
                    data=(row for row in batch),
                )

    # Finally, delete the original node (all children will also be deleted):
    delete_node(client, project_id, old_node_path)
    LOGGER.info(f"Successfully moved {old_node_path} to {new_node_path}")

    return



"""Tasks related to loading and using configuration files"""
from typing import Union, Optional
from pathlib import Path

import yaml
import logging
import re

import os

from tamr_toolbox.models.data_type import JsonDict

LOGGER = logging.getLogger(__name__)


# Define pattern to be used for parsing environment variables in config files
# This pattern searches for the '$' character at the start of variable name and then retrieves
# all other text in the variable name. For example: '$MY_VAR' matches and returns 'MY_VAR'
ENV_VAR_PATTERN = re.compile(r"^\$(.*)$")


def _yaml_env_variable_constructor(loader, node):
    """Replaces values that match the ENV_VAR_PATTERN with
    the value of the corresponding environment variable"""
    value = loader.construct_scalar(node)
    env_match = ENV_VAR_PATTERN.match(value)
    if env_match:
        value = env_match.groups()[0]
        env_var = os.environ.get(value)

        if env_var is None:
            LOGGER.warning(f"No environment variable for {value} found, using None.")
        return env_var
    return value


def _yaml_env_loader(path_to_file: Optional[Union[str, Path]]) -> JsonDict:
    """Reads a yaml file and creates a dictionary, retrieving environment variables as needed

    Args:
        path_to_file: Path to config yaml file

    Returns:
        All configuration variables in a dictionary
    """

    yaml_loader = yaml.SafeLoader
    yaml_loader.add_implicit_resolver("!env_var", ENV_VAR_PATTERN, None)
    yaml_loader.add_constructor("!env_var", _yaml_env_variable_constructor)

    with open(path_to_file, "r") as config_file:
        configs = yaml.load(config_file.read(), Loader=yaml_loader)
    LOGGER.info(f"Configurations have been loaded from {path_to_file}")
    return configs


def from_yaml(
    path_to_file: Optional[Union[str, Path]],
    *,
    default_path_to_file: Optional[Union[str, Path]] = None,
) -> JsonDict:
    """Reads a yaml file and creates a dictionary. Input values can be retrieved from environment
    variables

    Args:
        path_to_file: Path to config yaml file
        default_path_to_file: Path to use if path_to_file is null or empty
    Returns:
        All configuration variables in a dictionary
    """

    # If path_to_file is not populated, use default_path_to_file instead
    if path_to_file is None or (isinstance(path_to_file, str) and len(path_to_file) == 0):
        if default_path_to_file is None or (
            isinstance(default_path_to_file, str) and len(default_path_to_file) == 0
        ):
            not_populated_error = (
                f"No target path or default path has been specified. At least one must have a "
                f"value not: 'path_to_file'='{path_to_file}', "
                f"'default_path_to_file'='{default_path_to_file}'"
            )
            LOGGER.error(not_populated_error)
            raise ValueError(not_populated_error)
        else:
            LOGGER.info(f"Reading configuration file from default: {default_path_to_file}")
            path_to_file = default_path_to_file

    return _yaml_env_loader(path_to_file)



"""Tasks related to creating, updating, saving and moving translation dictionaries
in and out of Tamr"""
import json
import logging
import os
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Set, Union

from requests.exceptions import HTTPError
from tamr_unify_client.dataset.collection import DatasetCollection
from tamr_unify_client.dataset.resource import Dataset

from tamr_toolbox.enrichment.enrichment_utils import SetEncoder, create_empty_mapping

LOGGER = logging.getLogger(__name__)


@dataclass
class TranslationDictionary:
    """
    A DataClass for translation dictionaries

    Args:
        standardized_phrase: The unique common standardized version of all original_phrases
        translated_phrase: The translated standardized phrase to the target language of the
            dictionary
        detected_language: The language detected of the standardized phrase if source language is
            set to auto
        original_phrases: A set of original phrases which all convert to the standardized phrases
            when applying standardization
    """

    standardized_phrase: str = None
    translated_phrase: str = None
    detected_language: str = None
    original_phrases: Set[str] = field(default_factory=set)


def filename(
    dictionary_folder: Union[str, Path],
    *,
    target_language: str = "en",
    source_language: str = "auto",
) -> str:
    """
    Generate a toolbox translation dictionary file path

    Args:
        dictionary_folder: base directory where dictionaries are saved
        target_language: the language to translate into, for a list of allowed inputs:
            https://cloud.google.com/translate/docs/basic/discovering-supported-languages
        source_language: the language the text to translate is in, if None, assumes it is "auto"

    Returns:
        A toolbox translation dictionary file path

    """
    if source_language is None:
        source_language = "auto"
    dictionary_name = f"dictionary_{source_language.lower()}_to_{target_language.lower()}.json"
    if isinstance(dictionary_folder, str):
        dictionary_folder = Path(dictionary_folder)
    return str(dictionary_folder / dictionary_name)


def create(
    dictionary_folder: str, *, target_language: str = "en", source_language: str = "auto"
) -> str:
    """
    Create an empty dictionary on disk

    Args:
        dictionary_folder: base directory where dictionary is saved
        target_language: the language to translate into, for a list of allowed inputs:
            https://cloud.google.com/translate/docs/basic/discovering-supported-languages
        source_language: the language the text to translate is in, if None, assumes it is "auto"

    Returns:
        A path to a dictionary
    """
    dictionary_filepath = filename(
        dictionary_folder, target_language=target_language, source_language=source_language
    )
    create_empty_mapping(dictionary_filepath)
    return dictionary_filepath


def to_json(dictionary: Dict[str, TranslationDictionary]) -> List[str]:
    """
    Convert a toolbox translation dictionary entries to a json format where set object are
    converted to list

    Args:
        dictionary: a toolbox translation dictionary

    Returns:
        A list of toolbox translation dictionary entries in json format
    """
    return [json.dumps(asdict(t), cls=SetEncoder) for t in dictionary.values()]


def to_dict(dictionary: Dict[str, TranslationDictionary]) -> List[Dict[str, Union[str, List]]]:
    """
    Convert a toolbox translation dictionary entries to a dictionary format where set object are
    converted to list

    Args:
        dictionary: a toolbox translation dictionary

    Returns:
        A list of toolbox translation dictionary entries in dictionary format
    """
    return [json.loads(json.dumps(asdict(t), cls=SetEncoder)) for t in dictionary.values()]


def save(
    translation_dictionary: Dict[str, TranslationDictionary],
    dictionary_folder: str,
    *,
    target_language: str = "en",
    source_language: str = "auto",
) -> None:
    """
    Save a toolbox translation dictionary to disk

    Args:
        translation_dictionary: dictionary object to be saved to disk
        dictionary_folder: base directory where dictionary is saved
        target_language: the language to translate into, for a list of allowed inputs:
            https://cloud.google.com/translate/docs/basic/discovering-supported-languages
        source_language: the language the text to translate is in, if None, assumes it is "auto"

    Returns:

    """
    dictionary_filepath = filename(
        dictionary_folder, target_language=target_language, source_language=source_language
    )

    if len(translation_dictionary) > 0:
        LOGGER.debug("Writing Dictionary to file")
        with open(dictionary_filepath, "w") as f:
            f.write("\n".join(to_json(translation_dictionary)))


def load(
    dictionary_folder: str, *, target_language: str = "en", source_language: str = "auto"
) -> Dict[str, TranslationDictionary]:
    """
    Load a toolbox translation dictionary from disk to memory

    Args:
        dictionary_folder: base directory where dictionary is saved
        target_language: the language to translate into, for a list of allowed inputs:
            https://cloud.google.com/translate/docs/basic/discovering-supported-languages
        source_language: the language the text to translate is in, if None, assumes it is "auto"

    Returns:
        A toolbox translation dictionary

    Raises:
        RuntimeError: if the dictionary was found on disk but is not of a valid
            toolbox translation dictionary type
    """
    dictionary_filepath = filename(
        dictionary_folder, target_language=target_language, source_language=source_language
    )
    if not os.path.exists(dictionary_filepath):
        LOGGER.info(f"Dictionary {dictionary_filepath} does not exists, creating an empty one.")
        dictionary_filepath = create(
            dictionary_folder, target_language=target_language, source_language=source_language
        )

    with open(dictionary_filepath, "r") as f:
        translation_dictionary = [json.loads(line) for line in f.readlines()]
        try:
            # Tranform the loaded dictionaries into a TranslationDictionary
            translation_dictionary = [TranslationDictionary(**t) for t in translation_dictionary]
            # Change original phrases from List to Set
            for dictionary in translation_dictionary:
                dictionary.original_phrases = set(dictionary.original_phrases)
            # Make the standardized phrase the main key of the translation dictionary to be able
            # to access each translation easily
            translation_dictionary = {
                t.standardized_phrase: t
                for t in translation_dictionary
                if t.standardized_phrase is not None
            }
        except Exception as e:
            error_message = (
                f"Could not read translation dictionary at {dictionary_filepath}. "
                f"Check that the dictionary is of the correct type. Error: {e}"
            )
            LOGGER.error(error_message)
            raise RuntimeError(error_message)

    return translation_dictionary


def update(
    main_dictionary: Dict[str, TranslationDictionary],
    tmp_dictionary: Dict[str, TranslationDictionary],
) -> None:
    """
    Update a toolbox translation dictionary with another temporary translation dictionary

    Args:
        main_dictionary: the main toolbox translation dictionary containing past
            translation results
        tmp_dictionary: a temporary toolbox translation dictionary containing new translation
    Returns:

    """
    for standardized_phrase, translation in tmp_dictionary.items():
        try:
            main_dictionary_entry = main_dictionary[standardized_phrase]
            main_dictionary_entry.translated_phrase = translation.translated_phrase
            main_dictionary_entry.detected_language = translation.detected_language

        except KeyError:
            main_dictionary[standardized_phrase] = TranslationDictionary(
                standardized_phrase=standardized_phrase,
                translated_phrase=translation.translated_phrase,
                detected_language=translation.detected_language,
                original_phrases=translation.original_phrases,
            )


def convert_to_mappings(dictionary: Dict[str, TranslationDictionary]) -> Dict[str, str]:
    """
    Transform a translation dictionary into a mapping of original phrases to translated phrases
    Args:
        dictionary: a toolbox translation dictionary

    Returns:
        a dictionary with original phrase as key and translate phrase as value
    """
    mapping_from_dictionary = {
        original_phrase: t.translated_phrase
        for t in dictionary.values()
        for original_phrase in t.original_phrases
    }
    return mapping_from_dictionary


def from_dataset(dataset: Dataset) -> Dict[str, TranslationDictionary]:
    """
    Stream a dictionary from Tamr

    Args:
        dataset: Tamr Dataset object

    Returns:
        A toolbox translation dictionary

    Raises:
        ValueError: if the provided `dataset` is not a toolbox translation dictionary dataset
        NameError: if the provided `dataset` does not contain all the attributes of a
            toolbox translation dictionary
        RuntimeError: if there is any other problem while reading the `dataset` as a
            toolbox translation dictionary
    """
    if dataset.key_attribute_names[0] != "standardized_phrase":
        error_message = f"Provided Tamr Dataset is not a toolbox translation dictionary"
        LOGGER.error(error_message)
        raise ValueError(error_message)

    dictionary = {}
    for record in dataset.records():
        try:
            entry = TranslationDictionary(**record)
            # values are returned as a list of a single string, we change this to string
            entry.translated_phrase = entry.translated_phrase[0]
            entry.detected_language = entry.detected_language[0]

            # original phrases are stored on Tamr as lists, we save it as a set
            entry.original_phrases = set(entry.original_phrases)

        except NameError as e:
            error_message = (
                f"Supplied Tamr dataset is not in a toolbox translation dictionary format: {e}"
            )
            LOGGER.error(error_message)
            raise NameError(error_message)
        except Exception as e:
            error_message = f"Error while reading the Tamr dataset translation dictionary: {e}"
            LOGGER.error(error_message)
            raise RuntimeError(error_message)

        formatted_dictionary = {entry.standardized_phrase: entry}
        dictionary.update(formatted_dictionary)
    return dictionary


def to_dataset(
    dictionary: Dict[str, TranslationDictionary],
    *,
    dataset: Optional[Dataset] = None,
    datasets_collection: Optional[DatasetCollection] = None,
    target_language: Optional[str] = None,
    source_language: Optional[str] = None,
    create_dataset: bool = False,
) -> str:
    """
    Ingest a toolbox dictionary in Tamr, creates the source dataset if it doesn't exists

    Args:
        dictionary: a toolbox translation dictionary
        dataset: a Tamr client dataset
        datasets_collection: a Tamr client datasets collection
        target_language: the target language of the given dictionary
        source_language: the source language of the given dictionary
        create_dataset: flag to create or upsert to an existing translation dictionary
            source dataset

    Returns:
        The name of the created or updated Tamr Dataset

    Raises:
        ValueError: if `create_dataset` is False and `dataset` is not provided or is not a
            toolbox translation dictionary dataset.
            If `create_dataset` is True but `datasets_collection` or `target_language` or
            `source_language` is missing or the Tamr dataset already exists
        RuntimeError: if there is an error during the creation of the Tamr dataset attributes
    """
    if create_dataset is False:
        if dataset is None:
            error_message = (
                "Tamr Client Dataset missing from inputs, please provide a Tamr "
                "Client Dataset if updating an existing translation dictionary dataset"
            )
            LOGGER.error(error_message)
            raise ValueError(error_message)

        if dataset.key_attribute_names[0] != "standardized_phrase":
            error_message = f"Provided Tamr Dataset is not a toolbox translation dictionary"
            LOGGER.error(error_message)
            raise ValueError(error_message)

    else:
        if not (datasets_collection and target_language and source_language):
            error_message = (
                "A Tamr Datasets Collection, target_language and source_language "
                "must all be inputs if creating the toolbox translation "
                "dictionary dataset"
            )
            LOGGER.error(error_message)
            raise ValueError(error_message)

        # Get dataset name using filename function
        # The value of dictionary folder here is unimportant
        dataset_name = os.path.basename(
            filename(
                dictionary_folder="not/a/real/path",  # will be dropped immediately
                target_language=target_language,
                source_language=source_language,
            )
        )
        if dataset_name in [d.name for d in datasets_collection]:
            error_message = (
                f"Tamr Dataset {dataset_name} already exists on Tamr, you cannot "
                f"create a dataset with the same name as another one"
            )
            LOGGER.error(error_message)
            raise ValueError(error_message)

        LOGGER.info(f"Creating toolbox translation dictionary dataset {dataset_name} on Tamr")
        creation_spec = {"name": dataset_name, "keyAttributeNames": ["standardized_phrase"]}
        dataset = datasets_collection.create(creation_spec)

        attributes = dataset.attributes
        for attribute in ["translated_phrase", "detected_language", "original_phrases"]:
            attr_spec = {
                "name": attribute,
                "type": {"baseType": "ARRAY", "innerType": {"baseType": "STRING"}},
            }
            try:
                attributes.create(attr_spec)
            except HTTPError as e:
                error_message = (
                    f"Error while creating attribute {attribute} for dataset {dataset_name}: {e}"
                )
                LOGGER.error(error_message)
                raise RuntimeError(error_message)

    LOGGER.info("Ingesting toolbox translation dictionary to Tamr")
    dataset.upsert_records(records=to_dict(dictionary), primary_key_name="standardized_phrase")
    return dataset.name



# define Enum classes for project steps
from enum import Enum


class CategorizationSteps(Enum):
    """
    Enum class for the steps of a classification project
    """

    UPDATE_UNIFIED_DATASET = "updated_unified_dataset"
    APPLY_FEEDBACK = "apply_feedback"
    UPDATE_RESULTS_ONLY = "update_results_only"


class SchemaMappingSteps(Enum):
    """
    Enum class for the steps of a schema mapping project
    """

    UPDATE_UNIFIED_DATASET = "updated_unified_dataset"


class MasteringSteps(Enum):
    """
    Enum class for the steps of a mastering project
    """

    UPDATE_UNIFIED_DATASET = "update_unified_dataset"
    ESTIMATE_PAIR_COUNTS = "estimate_pair_counts"
    GENERATE_PAIRS = "generate_pairs"
    APPLY_FEEDBACK = "apply_feedback"
    UPDATE_HIGH_IMPACT_PAIRS = "update_high_impact_pairs"
    UPDATE_CLUSTERS = "update_clusters"
    APPLY_FEEDBACK_AND_UPDATE_RESULTS = "apply_feedback_and_update_results"
    UPDATE_RESULTS_ONLY = "update_results_only"
    PUBLISH_CLUSTERS = "publish_clusters"


class GoldenRecordsSteps(Enum):
    """
    Enum class for the steps of a golden records project
    """

    PROFILE_GOLDEN_RECORDS = "profile_golden_records"
    UPDATE_GOLDEN_RECORDS = "update_golden_records"
    PUBLISH_GOLDEN_RECORDS = "publish_golden_records"



"""
Helper functions related to creating & managing custom UI buttons as yaml files.

Due to how Tamr custom buttons are configured, these functions will
need to be run on the actual server on which Tamr is installed to work as expected.

Important: Custom buttons are only available to versions 2022.008.0 and later
"""
import logging
from typing import List, Optional, Union
import yaml
import os

from tamr_unify_client import Client
from tamr_toolbox.sysadmin.instance import _run_command
from tamr_toolbox.utils.version import requires_tamr_version

LOGGER = logging.getLogger(__name__)

TAMR_RELEASE_VERSION = "2022.008.0"


def _check_valid_page_name(*, pagename: str):
    """Checks that pagename provided is a valid Tamr button page name
    Args:
        pagename: Name of page provided

    Returns:
        True if page name is acceptable
    """
    valid_pagenames = [
        "Dataset Catalog",
        "Home",
        "Jobs",
        "Policies",
        "Users and Groups",
        "Categorization:Categorizations",
        "Categorization:Category Details",
        "Categorization:Dashboard",
        "Categorization:Project Datasets",
        "Categorization:Schema Mapping",
        "Categorization:Taxonomy",
        "Categorization:Unified Dataset",
        "Enrichment:Enrichment",
        "Golden Records:Golden Records",
        "Golden Records:Rules",
        "Mastering:Binning",
        "Mastering:Clusters",
        "Mastering:Dashboard",
        "Mastering:Group Records",
        "Mastering:Pairs",
        "Mastering:Project Datasets",
        "Mastering:Schema Mapping",
        "Mastering:Unified Dataset",
        "Schema Mapping:Dashboard",
        "Schema Mapping:Project Datasets",
        "Schema Mapping:Schema Mapping",
        "Schema Mapping:Unified Dataset",
    ]

    return pagename in valid_pagenames


def _check_valid_abs_path(dir: str):
    """
    Function to check provided path is absolute path
    and is not $TAMR_HOME/tamr/auxiliary-sevrices/conf

    Args:
        dir: directory provided

    Returns:

    Raises:
        ValueError: If path is not absolute or equal to
                    $TAMR_HOME/tamr/auxiliary-sevrices/conf
    """
    # Check absolute
    if not os.path.isabs(dir):
        value_error_message = f"Output directory must be absolute path."
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)

    bad_paths = ("tamr/auxiliary-sevrices/conf", "tamr/auxiliary-sevrices/conf/")

    if dir.endswith(bad_paths):
        value_error_message = (
            f"Output directory must not be $TAMR_HOME/tamr/auxiliary-sevrices/conf."
        )
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)


def create_redirect_button(
    *,
    extension_name: str,
    button_id: str,
    button_text: str,
    page_names: List[str],
    redirect_url: str,
    open_in_new_tab: bool,
    output_dir: str,
    button_name: str,
) -> str:
    """Create yaml file with all required attributes for
    a 'REDIRECT' UI button. Yaml file is saved locally.

    Button features are only available to versions 2022.008.0 and later.

    Args:
        extension_name: Name of button extension
        button_id: A short identifier for the button to use in the,
                   body of a POST call or a redirect URL path substitution.
        button_text: The button label to display in the UI.
        page_names: The pages of the UI on which to display the button.
        redirect_url: The URL that the browser should load
        open_in_new_tab: If true, the specified URL opens in a new browser tab.
        output_dir: Directory to save yaml file (absolute path)
        button_name: Name of yaml file

    Returns:
        Path to yaml file created
    """
    # Path validation
    _check_valid_abs_path(output_dir)

    # Minor url validation
    if not redirect_url.startswith(("http://", "https://")):
        value_error_message = f"Invalid url. Must begin with http:// or https://"
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)

    # Page name validation
    invalid_pages = [p for p in page_names if not _check_valid_page_name(pagename=p)]
    if invalid_pages:
        value_error_message = (
            f"Invalid pagename(s): {invalid_pages}. See docs for allowed Tamr button page names"
        )
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)

    button_dict = {
        "extensionName": extension_name,
        "buttons": [
            {
                "buttonType": "redirectButton",
                "buttonId": button_id,
                "buttonText": button_text,
                "pageNames": page_names,
                "redirectUrl": redirect_url,
                "openInNewTab": open_in_new_tab,
            }
        ],
    }

    file = f"{button_name}.yaml"
    filepath = os.path.join(output_dir, file)

    LOGGER.info(f"Saving {file} to {output_dir}")
    with open(f"{filepath}", "w") as yaml_file:
        yaml.dump(button_dict, yaml_file, sort_keys=False)

    return filepath


def create_post_button(
    *,
    extension_name: str,
    button_id: str,
    button_text: str,
    page_names: List[str],
    post_url: str,
    post_body_keys: List[str],
    success_message: str,
    fail_message: str,
    display_response: bool,
    output_dir: str,
    button_name: str,
) -> str:
    """
    Create yaml file with all required attributes for
    a 'POST' UI button. Yaml file is saved locally.

    Button features are only available to versions 2022.008.0 and later.

    Args:
        extension_name: Name of button extension
        button_id: A short identifier for the button to use in the,
                   body of a POST call or a redirect URL path substitution.
        button_text: The button label to display in the UI.
        page_names: The pages of the UI on which to display the button.
        post_url: The target URL for a POST API call
        post_body_keys: Specifies the keys to request in the body of the POST call
        success_message: The message that displays to the user when the POST call succeeds.
        fail_message: The message that displays to the user when the POST call fails.
        display_response: Whether the contents of the API response body should display to the user.
        output_dir: Directory to save yaml file (absolute path)
        button_name: Name of yaml file

    Returns:
        Path to yaml file created
    """
    # Path validation
    _check_valid_abs_path(output_dir)

    # Minor url validation
    if not post_url.startswith(("http://", "https://")):
        value_error_message = f"Invalid url. Must begin with http:// or https://"
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)

    # Page name validation
    invalid_pages = [p for p in page_names if not _check_valid_page_name(pagename=p)]
    if invalid_pages:
        value_error_message = (
            f"Invalid pagename(s): {invalid_pages}. See docs for allowed Tamr button page names"
        )
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)

    button_dict = {
        "extensionName": extension_name,
        "buttons": [
            {
                "buttonType": "postButton",
                "buttonId": button_id,
                "buttonText": button_text,
                "pageNames": page_names,
                "postUrl": post_url,
                "postBodyKeys": post_body_keys,
                "successMessage": success_message,
                "failMessage": fail_message,
                "displayResponse": display_response,
            }
        ],
    }

    file = f"{button_name}.yaml"
    filepath = os.path.join(output_dir, file)

    LOGGER.info(f"Saving {file} to {output_dir}")
    with open(f"{filepath}", "w") as yaml_file:
        yaml.dump(button_dict, yaml_file, sort_keys=False)

    return filepath


def create_button_extension(*, extension_name: str, buttons: List[str], output_dir: str) -> str:
    """
    Given a list of button yaml files, save it as a grouped extension yaml file.
    Yaml file is saved locally.
    Button features are only available to versions 2022.008.0 and later.

    Args:
        extension_name: Name of button extension to save
        buttons: List of button yaml files (absolute paths)
        output_dir: directory in which to save yaml extension file (absolute path)

    Returns:
        Path to yaml file created
    """
    # Path validation
    _check_valid_abs_path(output_dir)

    # Create dicts from the yaml files
    dict_list = []
    for file in buttons:
        with open(file, "r") as loaded_file:
            button_object = yaml.safe_load(loaded_file)
            button_dict = button_object["buttons"][0]
            dict_list.append(button_dict)

    file = f"{extension_name}.yaml"
    filepath = os.path.join(output_dir, file)

    output_dict = {"extensionName": extension_name, "buttons": dict_list}

    LOGGER.info(f"Saving {file} to {output_dir}")
    with open(f"{filepath}", "w") as yaml_file:
        yaml.dump(output_dict, yaml_file, sort_keys=False)

    return filepath


def create_button_extension_from_list(
    *, extension_name: str, output_dir: str, buttons: List[dict]
) -> str:
    """
    Given a list of button dictionaries, save it as a grouped extension yaml file.
    Yaml file is saved locally.

    Button features are only available to versions 2022.008.0 and later.

    Args:
        extension_name: Name of button extension to save
        output_dir: directory in which to save yaml extension file (absolute path)
        buttons: List of button dictionaries. Either redirect or post.
        Format examples:
        ---
        redirect:
        {
            "buttonType": "redirectButton",
            "buttonId": button_id,
            "buttonText": button_text,
            "pageNames": page_names,
            "redirectUrl": redirect_url,
            "openInNewTab": open_in_new_tab
        }
        ---
        post:
        {
            "buttonType": "postButton",
            "buttonId": button_id,
            "buttonText": button_text,
            "pageNames": page_names,
            "postUrl": post_url,
            "postBodyKeys": post_body_keys,
            "successMessage": success_message,
            "failMessage": fail_message,
            "displayResponse": display_response
        }
        ---

    Returns:
        Path to yaml file created
    """
    # Path Validation
    _check_valid_abs_path(output_dir)

    # URL & pagename validation
    invalid_urls = []
    invalid_pages = []
    for button_dict in buttons:
        if button_dict["buttonType"] == "postButton":
            url = button_dict["postUrl"]
        else:
            url = button_dict["redirectUrl"]

        if not url.startswith(("http://", "https://")):
            invalid_urls.append(url)

        invalid_pages += [
            p for p in button_dict["pageNames"] if not _check_valid_page_name(pagename=p)
        ]

    if invalid_urls and invalid_pages:
        value_error_message = f"Invalid url and pagenames. \
            url(s) {invalid_urls} must begin with http:// or https:// \
            invalid page name(s): {invalid_pages}"
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)
    elif invalid_urls and not invalid_pages:
        value_error_message = f"Invalid url(s) {invalid_urls}. Must begin with http:// or https://"
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)
    elif not invalid_urls and invalid_pages:
        value_error_message = (
            f"Invalid pagename(s): {invalid_pages}. See docs for allowed Tamr button page names"
        )
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)

    file = f"{extension_name}.yaml"
    filepath = os.path.join(output_dir, file)

    output_dict = {"extensionName": extension_name, "buttons": buttons}

    LOGGER.info(f"Saving {file} to {output_dir}")
    with open(f"{filepath}", "w") as yaml_file:
        yaml.dump(output_dict, yaml_file, sort_keys=False)

    return filepath


@requires_tamr_version(min_version=TAMR_RELEASE_VERSION)
def register_buttons(
    *,
    tamr_client: Client,
    buttons: Union[str, List[str]],
    tamr_install_dir: str,
    remote_client: Optional["paramiko.SSHClient"] = None,
    impersonation_username: Optional[str] = None,
    impersonation_password: Optional[str] = None,
):
    """
    Registers a list of button(s) in a Tamr instance.
    Requires Tamr restart to display buttons in UI.

    Important: If NOT running this function using a remote client, this function must
               be run on the server on which Tamr is installed.

    Runs in a remote environment if an ssh client is specified otherwise runs in the local shell.
    If an impersonation_username is provided, the command is run as the provided user.
    If an impersonation_password is provided, password authentication is used for impersonation,
    otherwise sudo is used.
    Button features are only available to versions 2022.008.0 and later.

    Version:
        Requires Tamr 2022.008.0 or later

    Args:
        tamr_client: Tamr Client object
        buttons: An individual string or a list of yaml files (absolute paths) with button configs
        tamr_install_dir: Full path to directory where Tamr is installed
        remote_client: An ssh client providing a remote connection
        impersonation_username: A bash user to run the command as,
            this should be the tamr install user
        impersonation_password: The password for the impersonation_username

    Returns:
    """
    if isinstance(buttons, str):
        buttons = [buttons]

    # Clean up path
    _check_valid_abs_path(tamr_install_dir)

    if tamr_install_dir.endswith("/"):
        tamr_install_dir = tamr_install_dir[:-1]

    LOGGER.info(f"Registering the following buttons in Tamr: {buttons}")

    for button in buttons:
        # Run the register command
        command = (
            f"{tamr_install_dir}/tamr/utils/unify-admin.sh ui:config --extensionConfig {button}"
        )

        LOGGER.info(f"Registering {button}")
        _run_command(
            command=command,
            remote_client=remote_client,
            impersonation_username=impersonation_username,
            impersonation_password=impersonation_password,
            enforce_success=True,
        )


def delete_buttons(*, button_files: Union[str, List[str]], tamr_install_dir: str):
    """Given a list of button yaml files, delete them thus removing the button from UI.

       NB: Registered buttons are located in $TAMR_HOME/tamr/auxiliary-sevrices/conf
           Requires restart of Tamr to register deletion.
           Button features are only available to versions 2022.008.0 and later.

    Args:
        button_files: Individual string or list of button yaml files (absolute paths)
        tamr_install_dir: Full path to directory where Tamr is installed (absolute path)
    Returns:
    """
    if isinstance(button_files, str):
        button_files = [button_files]

    # Check all files exist
    missing_files = [f for f in button_files if not os.path.exists(f)]
    if missing_files:
        warning_message = f"File(s) {missing_files} not found"
        LOGGER.warning(warning_message)

    # Work with present files only
    present_files = [x for x in button_files if x not in set(missing_files)]
    if not present_files:
        error_message = "None of provided files found."
        raise FileNotFoundError(error_message)

    button_dir = os.path.join(tamr_install_dir, "tamr/auxiliary-services/conf")

    # Check present button files provided reside in correct folder.
    path_list = present_files + [button_dir]

    if os.path.commonpath(path_list) != button_dir:
        value_error_message = f"All button files provided must belong to {button_dir} \
             otherwise deletion will not register."
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)

    LOGGER.info(f"Removing yaml files from {button_dir}")
    for file in present_files:
        os.remove(file)
        LOGGER.info(f"Deleted {file}")



"""Tasks common to moving data in and out of Tamr"""
from typing import Optional, List, Any, Iterable, Callable, Dict
import logging

from tamr_unify_client.dataset.resource import Dataset

LOGGER = logging.getLogger(__name__)


def _flatten_list(val: Any, *, delimiter: str = "|", force: bool = False) -> Any:
    """
    Flattens list types to strings.
    By default, only list[string] is flattened to a single string.
    The `force` parameter allows flattening of other inner types like numerics to a string.
    A list of length 1 becomes a string.
    A list with length longer than 1 has its elements concatenated with the specified `delimiter`.
    Non-list types are passed through directly.
    Intended to be applied using `applymap` to a DataFrame created from a Tamr Dataset,
    or to a pandas Series using `apply`.

    Args:
        val: value to convert
        delimiter: string to use as delimiter for concatenating lists
        force: if True, will force non-string inner types to string

    Returns:
        string or original value
    """
    if isinstance(val, list):
        if force:
            # convert inner values to strings.  f-strings are faster than str()
            return delimiter.join(f"{i}" for i in filter(None, val))
        else:
            return delimiter.join(filter(None, val))
    else:
        return val


def _yield_records(
    dataset: Dataset,
    *,
    func: Optional[Callable] = None,
    columns: Optional[List] = None,
    flatten_columns: Optional[List] = None,
) -> Iterable:
    """
    Generator function for records from a Tamr Dataset.
    Can optionally apply a flattening function to some or all columns,
    and can optionally keep only a subset of columns from the original records.

    Args:
        dataset: Tamr Dataset
        func: optional, callable function to transform records
        columns: optional, list of columns to keep
        flatten_columns: optional, list of columns to flatten

    Returns: iterable over Dataset records

    Raises:
        ValueError: if `flatten_columns` is not None but `func` is None since it
            won't know how to flatten them.
    """

    if flatten_columns is not None and func is None:
        message = (
            f"Columns specified for flatting ('flatten_columns'={flatten_columns}), "
            f"but no flattening function provided for parameter 'func'"
        )
        LOGGER.error(message)
        raise ValueError(message)

    checked_columns = False
    for record in dataset.records():
        if not checked_columns:
            if columns is not None:
                _check_columns_subset(
                    input_list=columns, reference_list=record.keys(), raise_error=True
                )
            if flatten_columns is not None:
                _check_columns_subset(
                    input_list=flatten_columns, reference_list=record.keys(), raise_error=True
                )
            checked_columns = True

        # Set flatten_columns to all if unspecified
        if func is not None and flatten_columns is None:
            flatten_columns = record.keys()

        output = {}
        for k, v in record.items():
            if columns is not None and k not in columns:
                # remove the column by skipping
                continue

            do_flatten_column = flatten_columns is not None and k in flatten_columns
            if do_flatten_column and func is not None:
                output[k] = func(v)
            else:
                output[k] = v
        yield output


def _check_columns_subset(
    *, input_list: List, reference_list: List, raise_error: bool = True
) -> bool:
    """
    Checks if `input_list` is a subset of `reference_list`.
    Returns a bool, or if `raise_error` is True, raises a ValueError on a failed check.

    Args:
        input_list: list to perform check for
        reference_list: reference list to check against for membership
        raise_error: if True, will raise a ValueError on failed check instead of returning

    Returns: bool

    Raises:
        ValueError: if `input_list` is not a subset of `reference_list`, and `raise_error` is True
    """
    input_set = set(input_list)
    reference_set = set(reference_list)
    if not input_set.issubset(reference_set):
        if raise_error:
            message = f"column(s) {input_set.difference(reference_set)} not found"
            LOGGER.error(message)
            raise ValueError(message)
        else:
            return False
    return True


def _get_column_mapping_dict(
    *,
    dataset_attribute_names: Iterable[str],
    column_name_dict: Dict[str, str] = None,
    columns: List[str] = None,
) -> Dict[str, str]:
    """
    Generate a dictionary to support column re-naming in CSV output.

    Args:
        dataset_attribute_names: Names of Tamr Dataset columns
        column_name_dict: Optional, Dictionary containing renaming information in the form
            {<Tamr dataset attribute name>: <new name>}. If None, preserve original names
        columns: Optional, List of columns to include in output. If None, use all columns
    Returns:
        Dictionary mapping Tamr Dataset attribute names to new column names
    Raises:
        ValueError: if column renaming would yield duplicate column names
    """

    if columns is None:
        columns = dataset_attribute_names

    if column_name_dict is not None:
        # Warn if the supplied column mapping has keys which don't appear in the dataset
        is_subset = _check_columns_subset(
            input_list=column_name_dict.keys(),
            reference_list=dataset_attribute_names,
            raise_error=False,
        )

        if not is_subset:
            message = (
                f"Some key column(s) in the column mapping {column_name_dict}"
                "do not appear in specified Tamr dataset."
            )
            LOGGER.warning(message)

        # Make mapping dict, mapping columns to themselves if they're not in the column_name_dict
        # Dict is generated by iterating through columns list so that order of columns is preserved
        full_column_name_dict = {col: column_name_dict.get(col, col) for col in columns}

        # Check that renaming won't generate duplicate columns in the output
        if len(set(full_column_name_dict.values())) < len(full_column_name_dict):
            message = "Column renaming would generate duplicate column names in output."
            LOGGER.error(message)
            raise ValueError(message)

    else:  # If there's no column renaming dictionary supplied, just map columns to themselves
        full_column_name_dict = dict(zip(columns, columns))

    return full_column_name_dict



"""Tasks related to Tamr Categorization projects"""
from . import jobs
from . import metrics
from . import taxonomy
from .._common import attribute_configuration, schema, transformations

__all__ = ["jobs", "attribute_configuration", "schema", "transformations", "metrics", "taxonomy"]



"""Enum and dataclasses representing attribute types in Tamr"""
from enum import Enum
from dataclasses import dataclass
from typing import ClassVar, Tuple, Union
from tamr_toolbox.models.data_type import JsonDict
from copy import deepcopy

PrimitiveType = Enum("PrimitiveType", ["BOOLEAN", "DOUBLE", "INT", "LONG", "STRING"])

# primitive type aliases
DOUBLE = PrimitiveType.DOUBLE
BOOLEAN = PrimitiveType.BOOLEAN
INT = PrimitiveType.INT
LONG = PrimitiveType.LONG
STRING = PrimitiveType.STRING


@dataclass(frozen=True)
class Array:
    """See https://docs.tamr.com/reference#attribute-types
    NOTE:
        `sphinx_autodoc_typehints` cannot handle forward reference to `AttributeType`,
        so reference docs are written manually for this type
    Args:
        inner_type
    """

    _tag: ClassVar[str] = "ARRAY"
    inner_type: "AttributeType"


@dataclass(frozen=True)
class Map:
    """See https://docs.tamr.com/reference#attribute-types
    NOTE:
        `sphinx_autodoc_typehints` cannot handle forward reference to `AttributeType`,
        so reference docs are written manually for this type
    Args:
        inner_type
    """

    _tag: ClassVar[str] = "MAP"
    inner_type: "AttributeType"


@dataclass(frozen=True)
class SubAttribute:
    """An attribute which is itself a property of another attribute.
    See https://docs.tamr.com/reference#attribute-types
    NOTE:
        `sphinx_autodoc_typehints` cannot handle forward reference to `AttributeType`,
        so reference docs are written manually for this type
    Args:
        name: Name of sub-attribute
        type: See https://docs.tamr.com/reference#attribute-types
        is_nullable: If this sub-attribute can be null
    """

    name: str
    type: "AttributeType"
    is_nullable: bool


@dataclass(frozen=True)
class Record:
    """See https://docs.tamr.com/reference#attribute-types
    Args:
        attributes
    """

    _tag: ClassVar[str] = "RECORD"
    attributes: Tuple[SubAttribute, ...]


ComplexType = Union[Array, Map, Record]

AttributeType = Union[PrimitiveType, ComplexType]

# complex type aliases
DEFAULT: AttributeType = Array(STRING)
GEOSPATIAL: AttributeType = Record(
    attributes=(
        SubAttribute(name="point", is_nullable=True, type=Array(DOUBLE)),
        SubAttribute(name="multiPoint", is_nullable=True, type=Array(Array(DOUBLE))),
        SubAttribute(name="lineString", is_nullable=True, type=Array(Array(DOUBLE))),
        SubAttribute(name="multiLineString", is_nullable=True, type=Array(Array(Array(DOUBLE)))),
        SubAttribute(name="polygon", is_nullable=True, type=Array(Array(Array(DOUBLE)))),
        SubAttribute(
            name="multiPolygon", is_nullable=True, type=Array(Array(Array(Array(DOUBLE))))
        ),
    )
)


def from_json(data: JsonDict) -> AttributeType:
    """Make an attribute type from JSON data (deserialize)
    Args:
        data: JSON data from Tamr server
    """
    base_type = data.get("baseType")
    if base_type is None:
        raise ValueError("Missing required field 'baseType'.")

    for primitive in PrimitiveType:
        if base_type == primitive.name:
            return primitive

    if base_type == Array._tag:
        inner_type = data.get("innerType")
        if inner_type is None:
            raise ValueError("Missing required field 'innerType' for Array type.")
        return Array(inner_type=from_json(inner_type))
    elif base_type == Map._tag:
        inner_type = data.get("innerType")
        if inner_type is None:
            raise ValueError("Missing required field 'innerType' for Map type.")
        return Map(inner_type=from_json(inner_type))
    elif base_type == Record._tag:
        attributes = data.get("attributes")
        if attributes is None:
            raise ValueError("Missing required field 'attributes' for Record type.")
        return Record(attributes=tuple([_subattribute_from_json(attr) for attr in attributes]))
    else:
        raise ValueError(f"Unrecognized 'baseType': {base_type}")


def _subattribute_from_json(data: JsonDict) -> SubAttribute:
    """Make a SubAttribute from JSON data (deserialize)
    Args:
        data: JSON data received from Tamr server.
    """

    cp = deepcopy(data)
    d = {}
    d["name"] = cp["name"]
    d["is_nullable"] = cp["isNullable"]
    d["type"] = from_json(cp["type"])
    return SubAttribute(**d)


def to_json(attr_type: AttributeType) -> JsonDict:
    """Serialize attribute type to JSON
    Args:
        attr_type: Attribute type to serialize
    """
    if isinstance(attr_type, PrimitiveType):
        return {"baseType": attr_type.name}
    elif isinstance(attr_type, (Array, Map)):
        return {"baseType": type(attr_type)._tag, "innerType": to_json(attr_type.inner_type)}
    elif isinstance(attr_type, Record):

        return {
            "baseType": type(attr_type)._tag,
            "attributes": [
                {
                    "name": subattr.name,
                    "type": to_json(subattr.type),
                    "isNullable": subattr.is_nullable,
                }
                for subattr in attr_type.attributes
            ],
        }
    else:
        raise TypeError(attr_type)
